from enum import Enum
from typing import Optional

from xsdata_pydantic.fields import field

from libresvip.model.base import BaseModel

from .enums import VocaloidLanguage

VSQ4_NS = "http://www.yamaha.co.jp/vocaloid/schema/vsq4/"


class ParameterNames(Enum):
    DYN = "D"
    BRE = "B"
    BRI = "R"
    CLE = "C"
    GEN = "G"
    POR = "T"
    PBS = "S"
    PIT = "P"
    XSY = "X"
    GWL = "W"


class Aux(BaseModel):
    class Meta:
        name = "aux"
        namespace = VSQ4_NS

    aux_id: str | None = field(
        default=None,
        metadata={
            "name": "id",
            "type": "Element",
            "required": True,
        },
    )
    content: str | None = field(
        default=None,
        metadata={
            "type": "Element",
            "required": True,
        },
    )


class StylePlugin(BaseModel):
    class Meta:
        name = "sPlug"
        namespace = VSQ4_NS

    style_plugin_id: str | None = field(
        default="ACA9C502-A04B-42b5-B2EB-5CEA36D16FCE",
        metadata={
            "name": "id",
            "type": "Element",
            "required": True,
            "length": 36,
        },
    )
    style_plugin_name: str | None = field(
        default="VOCALOID2 Compatible Style",
        metadata={
            "name": "name",
            "type": "Element",
            "required": True,
        },
    )
    version: str | None = field(
        default="3.0.0.1",
        metadata={
            "type": "Element",
            "required": True,
        },
    )


class SeqAttr(BaseModel):
    class Meta:
        name = "seq"
        namespace = VSQ4_NS

    elem: list["SeqAttr.Elem"] = field(
        default_factory=list,
        metadata={
            "name": "cc",
            "type": "Element",
            "min_occurs": 1,
        },
    )
    seq_id: str | None = field(
        default=None,
        metadata={
            "name": "id",
            "type": "Attribute",
            "required": True,
        },
    )

    class Elem(BaseModel):
        pos_nrm: int | None = field(
            default=None,
            metadata={
                "name": "p",
                "type": "Element",
                "required": True,
            },
        )
        elv: int | None = field(
            default=None,
            metadata={
                "name": "v",
                "type": "Element",
                "required": True,
            },
        )


class Singer(BaseModel):
    class Meta:
        name = "singer"
        namespace = VSQ4_NS

    pos_tick: int | None = field(
        default=0,
        metadata={
            "name": "t",
            "type": "Element",
            "required": True,
        },
    )
    v_bs: VocaloidLanguage | None = field(
        default=VocaloidLanguage.SIMPLIFIED_CHINESE,
        metadata={
            "name": "bs",
            "type": "Element",
            "required": True,
            "min_inclusive": 0,
            "max_inclusive": 127,
        },
    )
    v_pc: int | None = field(
        default=0,
        metadata={
            "name": "pc",
            "type": "Element",
            "required": True,
            "min_inclusive": 0,
            "max_inclusive": 127,
        },
    )


class Tempo(BaseModel):
    class Meta:
        name = "tempo"
        namespace = VSQ4_NS

    pos_tick: int = field(
        metadata={
            "name": "t",
            "type": "Element",
            "required": True,
        },
    )
    bpm: int = field(
        metadata={
            "name": "v",
            "type": "Element",
            "required": True,
        },
    )


class TimeSig(BaseModel):
    class Meta:
        name = "timeSig"
        namespace = VSQ4_NS

    pos_mes: int | None = field(
        default=None,
        metadata={
            "name": "m",
            "type": "Element",
            "required": True,
        },
    )
    nume: int | None = field(
        default=None,
        metadata={
            "name": "nu",
            "type": "Element",
            "required": True,
        },
    )
    denomi: int | None = field(
        default=None,
        metadata={
            "name": "de",
            "type": "Element",
            "required": True,
        },
    )


class TypeParamAttr(BaseModel):
    class Meta:
        name = "typeParamAttr"

    value: int | None = field(
        default=None,
        metadata={
            "required": True,
        },
    )
    type_param_attr_id: str | None = field(
        default=None,
        metadata={
            "name": "id",
            "type": "Attribute",
            "required": True,
        },
    )


class TypePhonemes(BaseModel):
    class Meta:
        name = "typePhonemes"

    value: str = field(
        default="",
        metadata={
            "required": True,
        },
    )
    lock: int | None = field(
        default=None,
        metadata={
            "type": "Attribute",
            "min_inclusive": 0,
            "max_inclusive": 1,
        },
    )


class TypeResolution(Enum):
    VALUE_16 = 16
    VALUE_24 = 24
    VALUE_32 = 32


class TypeSamplerate(Enum):
    VALUE_44100 = 44100
    VALUE_48000 = 48000
    VALUE_96000 = 96000


class TypeVstSdkversion(Enum):
    VALUE_0 = 0
    VALUE_2 = 2
    VALUE_3 = 3


class VVoiceParam(BaseModel):
    class Meta:
        name = "vPrm"
        namespace = VSQ4_NS

    bre: int | None = field(
        default=0,
        metadata={
            "type": "Element",
            "required": True,
            "min_inclusive": -127,
            "max_inclusive": 127,
        },
    )
    bri: int | None = field(
        default=0,
        metadata={
            "type": "Element",
            "required": True,
            "min_inclusive": -127,
            "max_inclusive": 127,
        },
    )
    cle: int | None = field(
        default=0,
        metadata={
            "type": "Element",
            "required": True,
            "min_inclusive": -127,
            "max_inclusive": 127,
        },
    )
    gen: int | None = field(
        default=0,
        metadata={
            "type": "Element",
            "required": True,
            "min_inclusive": -127,
            "max_inclusive": 127,
        },
    )
    ope: int | None = field(
        default=0,
        metadata={
            "type": "Element",
            "required": True,
            "min_inclusive": -127,
            "max_inclusive": 127,
        },
    )


class VPrm2(BaseModel):
    class Meta:
        name = "vPrm2"
        namespace = VSQ4_NS

    bre: int | None = field(
        default=None,
        metadata={
            "type": "Element",
            "required": True,
            "min_inclusive": -127,
            "max_inclusive": 127,
        },
    )
    bri: int | None = field(
        default=None,
        metadata={
            "type": "Element",
            "required": True,
            "min_inclusive": -127,
            "max_inclusive": 127,
        },
    )
    cle: int | None = field(
        default=None,
        metadata={
            "type": "Element",
            "required": True,
            "min_inclusive": -127,
            "max_inclusive": 127,
        },
    )
    gen: int | None = field(
        default=None,
        metadata={
            "type": "Element",
            "required": True,
            "min_inclusive": -127,
            "max_inclusive": 127,
        },
    )
    ope: int | None = field(
        default=None,
        metadata={
            "type": "Element",
            "required": True,
            "min_inclusive": -127,
            "max_inclusive": 127,
        },
    )
    vol: int | None = field(
        default=None,
        metadata={
            "type": "Element",
            "required": True,
            "min_inclusive": -127,
            "max_inclusive": 127,
        },
    )


class MCtrl(BaseModel):
    class Meta:
        name = "cc"
        namespace = VSQ4_NS

    pos_tick: int = field(
        metadata={
            "name": "t",
            "type": "Element",
            "required": True,
        },
    )
    attr: TypeParamAttr = field(
        default_factory=TypeParamAttr,
        metadata={
            "name": "v",
            "type": "Element",
            "required": True,
        },
    )


class MasterTrack(BaseModel):
    class Meta:
        name = "masterTrack"
        namespace = VSQ4_NS

    seq_name: str | None = field(
        default="Untitled0",
        metadata={
            "name": "seqName",
            "type": "Element",
            "required": True,
        },
    )
    comment: str | None = field(
        default="New VSQ File",
        metadata={
            "type": "Element",
            "required": True,
        },
    )
    resolution: int | None = field(
        default=480,
        metadata={
            "type": "Element",
            "required": True,
        },
    )
    pre_measure: int = field(
        default=1,
        metadata={
            "name": "preMeasure",
            "type": "Element",
            "required": True,
            "min_inclusive": 0,
            "max_inclusive": 127,
        },
    )
    time_sig: list[TimeSig] = field(
        default_factory=list,
        metadata={
            "name": "timeSig",
            "type": "Element",
            "min_occurs": 1,
        },
    )
    tempo: list[Tempo] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "min_occurs": 1,
        },
    )


class NoteStyle(BaseModel):
    class Meta:
        name = "nStyle"
        namespace = VSQ4_NS

    attr: list[TypeParamAttr] = field(
        default_factory=list,
        metadata={
            "name": "v",
            "type": "Element",
        },
    )
    seq_attr: list[SeqAttr] = field(
        default_factory=list,
        metadata={
            "name": "seq",
            "type": "Element",
        },
    )


class PartStyle(BaseModel):
    class Meta:
        name = "pStyle"
        namespace = VSQ4_NS

    attr: list[TypeParamAttr] = field(
        default_factory=list,
        metadata={
            "name": "v",
            "type": "Element",
            "min_occurs": 1,
        },
    )


class VstPlugin(BaseModel):
    class Meta:
        name = "plug"
        namespace = VSQ4_NS

    vst_plugin_id: str | None = field(
        default=None,
        metadata={
            "name": "id",
            "type": "Element",
            "required": True,
        },
    )
    vst_plugin_name: str | None = field(
        default=None,
        metadata={
            "name": "name",
            "type": "Element",
            "required": True,
        },
    )
    vst_sdkversion: TypeVstSdkversion | None = field(
        default=None,
        metadata={
            "name": "sdkVer",
            "type": "Element",
            "required": True,
        },
    )
    vst_param_num: int | None = field(
        default=None,
        metadata={
            "name": "nPrm",
            "type": "Element",
            "required": True,
        },
    )
    vst_param_val: Optional["VstPlugin.VstParamVal"] = field(
        default=None,
        metadata={
            "name": "vPrm",
            "type": "Element",
        },
    )
    vst_preset_no: int | None = field(
        default=None,
        metadata={
            "name": "presetNo",
            "type": "Element",
        },
    )
    enable: int | None = field(
        default=None,
        metadata={
            "type": "Element",
            "required": True,
            "min_inclusive": 0,
            "max_inclusive": 1,
        },
    )
    bypass: int | None = field(
        default=None,
        metadata={
            "type": "Element",
            "required": True,
            "min_inclusive": 0,
            "max_inclusive": 1,
        },
    )

    class VstParamVal(BaseModel):
        val: list[int] = field(
            default_factory=list,
            metadata={
                "name": "v",
                "type": "Element",
            },
        )


class VstPluginSr(BaseModel):
    class Meta:
        name = "plugSR"
        namespace = VSQ4_NS

    vst_plugin_id: str | None = field(
        default=None,
        metadata={
            "name": "id",
            "type": "Element",
            "required": True,
        },
    )
    vst_plugin_name: str | None = field(
        default=None,
        metadata={
            "name": "name",
            "type": "Element",
            "required": True,
        },
    )
    vst_sdkversion: TypeVstSdkversion | None = field(
        default=None,
        metadata={
            "name": "sdkVer",
            "type": "Element",
            "required": True,
        },
    )
    vst_param_num: int | None = field(
        default=None,
        metadata={
            "name": "nPrm",
            "type": "Element",
            "required": True,
        },
    )
    vst_param_val: Optional["VstPluginSr.VstParamVal"] = field(
        default=None,
        metadata={
            "name": "vPrm",
            "type": "Element",
        },
    )
    vst_preset_no: int | None = field(
        default=None,
        metadata={
            "name": "presetNo",
            "type": "Element",
        },
    )
    enable: int | None = field(
        default=None,
        metadata={
            "type": "Element",
            "required": True,
            "min_inclusive": 0,
            "max_inclusive": 1,
        },
    )
    bypass: int | None = field(
        default=None,
        metadata={
            "type": "Element",
            "required": True,
            "min_inclusive": 0,
            "max_inclusive": 1,
        },
    )

    class VstParamVal(BaseModel):
        val: list[int] = field(
            default_factory=list,
            metadata={
                "name": "v",
                "type": "Element",
            },
        )


class VVoice(BaseModel):
    class Meta:
        name = "vVoice"
        namespace = VSQ4_NS

    v_bs: VocaloidLanguage | None = field(
        default=VocaloidLanguage.SIMPLIFIED_CHINESE,
        metadata={
            "name": "bs",
            "type": "Element",
            "required": True,
            "min_inclusive": 0,
            "max_inclusive": 127,
        },
    )
    v_pc: int | None = field(
        default=0,
        metadata={
            "name": "pc",
            "type": "Element",
            "required": True,
            "min_inclusive": 0,
            "max_inclusive": 127,
        },
    )
    comp_id: str | None = field(
        default=None,
        metadata={
            "name": "id",
            "type": "Element",
            "required": True,
            "length": 16,
        },
    )
    v_voice_name: str | None = field(
        default=None,
        metadata={
            "name": "name",
            "type": "Element",
            "required": True,
        },
    )
    v_voice_param: VVoiceParam | None = field(
        default_factory=VVoiceParam,
        metadata={
            "name": "vPrm",
            "type": "Element",
            "required": True,
        },
    )
    id2: str | None = field(
        default=None,
        metadata={
            "type": "Element",
            "length": 16,
        },
    )
    v_prm2: VPrm2 | None = field(
        default=None,
        metadata={
            "name": "vPrm2",
            "type": "Element",
        },
    )


class WavPart(BaseModel):
    class Meta:
        name = "wavPart"
        namespace = VSQ4_NS

    pos_tick: int | None = field(
        default=None,
        metadata={
            "name": "t",
            "type": "Element",
            "required": True,
        },
    )
    play_time: int | None = field(
        default=None,
        metadata={
            "name": "playTime",
            "type": "Element",
            "required": True,
        },
    )
    part_name: str | None = field(
        default=None,
        metadata={
            "name": "name",
            "type": "Element",
            "required": True,
        },
    )
    comment: str | None = field(
        default="New Wav Part",
        metadata={
            "type": "Element",
            "required": True,
        },
    )
    sample_rate: TypeSamplerate | None = field(
        default=None,
        metadata={
            "name": "fs",
            "type": "Element",
            "required": True,
        },
    )
    sample_reso: TypeResolution | None = field(
        default=None,
        metadata={
            "name": "rs",
            "type": "Element",
            "required": True,
        },
    )
    channels: int | None = field(
        default=None,
        metadata={
            "name": "nCh",
            "type": "Element",
            "required": True,
            "min_inclusive": 1,
            "max_inclusive": 2,
        },
    )
    file_path: str | None = field(
        default=None,
        metadata={
            "name": "filePath",
            "type": "Element",
            "required": True,
        },
    )


class MasterUnit(BaseModel):
    class Meta:
        name = "masterUnit"
        namespace = VSQ4_NS

    out_dev: int | None = field(
        default=0,
        metadata={
            "name": "oDev",
            "type": "Element",
            "required": True,
        },
    )
    vst_plugin: list[VstPlugin] = field(
        default_factory=list,
        metadata={
            "name": "plug",
            "type": "Element",
            "max_occurs": 2,
        },
    )
    vst_plugin_sr: VstPluginSr | None = field(
        default=None,
        metadata={
            "name": "plugSR",
            "type": "Element",
        },
    )
    ret_level: int | None = field(
        default=0,
        metadata={
            "name": "rLvl",
            "type": "Element",
            "required": True,
        },
    )
    vol: int | None = field(
        default=0,
        metadata={
            "type": "Element",
            "required": True,
        },
    )


class MonoTrack(BaseModel):
    class Meta:
        name = "monoTrack"
        namespace = VSQ4_NS

    wav_part: list[WavPart] = field(
        default_factory=list,
        metadata={
            "name": "wavPart",
            "type": "Element",
        },
    )


class MonoUnit(BaseModel):
    class Meta:
        name = "monoUnit"
        namespace = VSQ4_NS

    in_gain: int | None = field(
        default=0,
        metadata={
            "name": "iGin",
            "type": "Element",
            "required": True,
        },
    )
    vst_plugin: list[VstPlugin] = field(
        default_factory=list,
        metadata={
            "name": "plug",
            "type": "Element",
            "max_occurs": 2,
        },
    )
    send_level: int | None = field(
        default=-898,
        metadata={
            "name": "sLvl",
            "type": "Element",
            "required": True,
        },
    )
    send_enable: int | None = field(
        default=0,
        metadata={
            "name": "sEnable",
            "type": "Element",
            "required": True,
            "min_inclusive": 0,
            "max_inclusive": 1,
        },
    )
    mute: int | None = field(
        default=0,
        metadata={
            "name": "m",
            "type": "Element",
            "required": True,
            "min_inclusive": 0,
            "max_inclusive": 1,
        },
    )
    solo: int | None = field(
        default=0,
        metadata={
            "name": "s",
            "type": "Element",
            "required": True,
            "min_inclusive": 0,
            "max_inclusive": 1,
        },
    )
    pan: int | None = field(
        default=64,
        metadata={
            "type": "Element",
            "required": True,
        },
    )
    vol: int | None = field(
        default=0,
        metadata={
            "type": "Element",
            "required": True,
        },
    )


class Note(BaseModel):
    class Meta:
        name = "note"
        namespace = VSQ4_NS

    pos_tick: int = field(
        metadata={
            "name": "t",
            "type": "Element",
            "required": True,
        },
    )
    dur_tick: int = field(
        metadata={
            "name": "dur",
            "type": "Element",
            "required": True,
        },
    )
    note_num: int | None = field(
        default=None,
        metadata={
            "name": "n",
            "type": "Element",
            "required": True,
            "min_inclusive": 0,
            "max_inclusive": 127,
        },
    )
    velocity: int | None = field(
        default=64,
        metadata={
            "name": "v",
            "type": "Element",
            "required": True,
            "min_inclusive": 0,
            "max_inclusive": 127,
        },
    )
    lyric: str = field(
        metadata={
            "name": "y",
            "type": "Element",
            "required": True,
        },
    )
    phnms: TypePhonemes | None = field(
        default=None,
        metadata={
            "name": "p",
            "type": "Element",
            "required": True,
        },
    )
    note_style: NoteStyle = field(
        default_factory=NoteStyle,
        metadata={
            "name": "nStyle",
            "type": "Element",
        },
    )


class StereoTrack(BaseModel):
    class Meta:
        name = "stTrack"
        namespace = VSQ4_NS

    wav_part: list[WavPart] = field(
        default_factory=list,
        metadata={
            "name": "wavPart",
            "type": "Element",
        },
    )


class StereoUnit(BaseModel):
    class Meta:
        name = "stUnit"
        namespace = VSQ4_NS

    in_gain: int | None = field(
        default=0,
        metadata={
            "name": "iGin",
            "type": "Element",
            "required": True,
        },
    )
    vst_plugin: list[VstPlugin] = field(
        default_factory=list,
        metadata={
            "name": "plug",
            "type": "Element",
            "max_occurs": 2,
        },
    )
    mute: int | None = field(
        default=0,
        metadata={
            "name": "m",
            "type": "Element",
            "required": True,
            "min_inclusive": 0,
            "max_inclusive": 1,
        },
    )
    solo: int | None = field(
        default=0,
        metadata={
            "name": "s",
            "type": "Element",
            "required": True,
            "min_inclusive": 0,
            "max_inclusive": 1,
        },
    )
    vol: int | None = field(
        default=0,
        metadata={
            "type": "Element",
            "required": True,
        },
    )


class VVoiceTable(BaseModel):
    class Meta:
        name = "vVoiceTable"
        namespace = VSQ4_NS

    v_voice: list[VVoice] = field(
        default_factory=list,
        metadata={
            "name": "vVoice",
            "type": "Element",
            "min_occurs": 1,
        },
    )


class VsUnit(BaseModel):
    class Meta:
        name = "vsUnit"
        namespace = VSQ4_NS

    vs_track_no: int | None = field(
        default=None,
        metadata={
            "name": "tNo",
            "type": "Element",
            "required": True,
            "min_inclusive": 0,
            "max_inclusive": 127,
        },
    )
    in_gain: int | None = field(
        default=0,
        metadata={
            "name": "iGin",
            "type": "Element",
            "required": True,
        },
    )
    vst_plugin: list[VstPlugin] = field(
        default_factory=list,
        metadata={
            "name": "plug",
            "type": "Element",
            "max_occurs": 2,
        },
    )
    send_level: int | None = field(
        default=-898,
        metadata={
            "name": "sLvl",
            "type": "Element",
            "required": True,
        },
    )
    send_enable: int | None = field(
        default=0,
        metadata={
            "name": "sEnable",
            "type": "Element",
            "required": True,
            "min_inclusive": 0,
            "max_inclusive": 1,
        },
    )
    mute: int | None = field(
        default=0,
        metadata={
            "name": "m",
            "type": "Element",
            "required": True,
            "min_inclusive": 0,
            "max_inclusive": 1,
        },
    )
    solo: int | None = field(
        default=0,
        metadata={
            "name": "s",
            "type": "Element",
            "required": True,
            "min_inclusive": 0,
            "max_inclusive": 1,
        },
    )
    pan: int | None = field(
        default=64,
        metadata={
            "type": "Element",
            "required": True,
        },
    )
    vol: int | None = field(
        default=0,
        metadata={
            "type": "Element",
            "required": True,
        },
    )


class Mixer(BaseModel):
    class Meta:
        name = "mixer"
        namespace = VSQ4_NS

    master_unit: MasterUnit | None = field(
        default_factory=MasterUnit,
        metadata={
            "name": "masterUnit",
            "type": "Element",
            "required": True,
        },
    )
    vs_unit: list[VsUnit] = field(
        default_factory=list,
        metadata={
            "name": "vsUnit",
            "type": "Element",
            "min_occurs": 1,
            "max_occurs": 16,
        },
    )
    mono_unit: MonoUnit = field(
        default_factory=MonoUnit,
        metadata={
            "name": "monoUnit",
            "type": "Element",
            "required": True,
        },
    )
    stereo_unit: StereoUnit = field(
        default_factory=StereoUnit,
        metadata={
            "name": "stUnit",
            "type": "Element",
            "required": True,
        },
    )


class MusicalPart(BaseModel):
    class Meta:
        name = "vsPart"
        namespace = VSQ4_NS

    pos_tick: int = field(
        metadata={
            "name": "t",
            "type": "Element",
            "required": True,
        },
    )
    play_time: int = field(
        metadata={
            "name": "playTime",
            "type": "Element",
            "required": True,
        },
    )
    part_name: str | None = field(
        default="New Part",
        metadata={
            "name": "name",
            "type": "Element",
            "required": True,
        },
    )
    comment: str | None = field(
        default="New Musical Part",
        metadata={
            "type": "Element",
            "required": True,
        },
    )
    style_plugin: StylePlugin | None = field(
        default_factory=StylePlugin,
        metadata={
            "name": "sPlug",
            "type": "Element",
            "required": True,
        },
    )
    part_style: PartStyle = field(
        default_factory=PartStyle,
        metadata={
            "name": "pStyle",
            "type": "Element",
        },
    )
    singer: list[Singer] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "min_occurs": 1,
        },
    )
    m_ctrl: list[MCtrl] = field(
        default_factory=list,
        metadata={
            "name": "cc",
            "type": "Element",
        },
    )
    note: list[Note] = field(
        default_factory=list,
        metadata={
            "type": "Element",
        },
    )
    plane: int | None = field(
        default=None,
        metadata={
            "type": "Element",
            "min_inclusive": 0,
            "max_inclusive": 1,
        },
    )


class VsTrack(BaseModel):
    class Meta:
        name = "vsTrack"
        namespace = VSQ4_NS

    vs_track_no: int | None = field(
        default=None,
        metadata={
            "name": "tNo",
            "type": "Element",
            "required": True,
            "min_inclusive": 0,
            "max_inclusive": 127,
        },
    )
    track_name: str | None = field(
        default=None,
        metadata={
            "name": "name",
            "type": "Element",
            "required": True,
        },
    )
    comment: str | None = field(
        default="Track",
        metadata={
            "type": "Element",
            "required": True,
        },
    )
    musical_part: list[MusicalPart] = field(
        default_factory=list,
        metadata={
            "name": "vsPart",
            "type": "Element",
        },
    )


class Vsq4(BaseModel):
    class Meta:
        name = "vsq4"
        namespace = VSQ4_NS

    vender: str | None = field(
        default="Yamaha Corporation",
        metadata={
            "type": "Element",
            "required": True,
        },
    )
    version: str | None = field(
        default="4.0.0.3",
        metadata={
            "type": "Element",
            "required": True,
        },
    )
    v_voice_table: VVoiceTable = field(
        default_factory=VVoiceTable,
        metadata={
            "name": "vVoiceTable",
            "type": "Element",
            "required": True,
        },
    )
    mixer: Mixer = field(
        default_factory=Mixer,
        metadata={
            "type": "Element",
            "required": True,
        },
    )
    master_track: MasterTrack = field(
        default_factory=MasterTrack,
        metadata={
            "name": "masterTrack",
            "type": "Element",
            "required": True,
        },
    )
    vs_track: list[VsTrack] = field(
        default_factory=list,
        metadata={
            "name": "vsTrack",
            "type": "Element",
            "min_occurs": 1,
            "max_occurs": 16,
        },
    )
    mono_track: MonoTrack | None = field(
        default_factory=MonoTrack,
        metadata={
            "name": "monoTrack",
            "type": "Element",
            "required": True,
        },
    )
    stereo_track: StereoTrack | None = field(
        default_factory=StereoTrack,
        metadata={
            "name": "stTrack",
            "type": "Element",
            "required": True,
        },
    )
    aux: list[Aux] = field(
        default_factory=list,
        metadata={
            "type": "Element",
        },
    )



import dataclasses
import operator
import re
from collections.abc import Callable
from functools import partial, reduce
from typing import cast

from libresvip.core.constants import DEFAULT_BPM
from libresvip.core.tick_counter import shift_beat_list, shift_tempo_list
from libresvip.core.time_interval import RangeInterval
from libresvip.core.time_sync import TimeSynchronizer
from libresvip.model.base import (
    InstrumentalTrack,
    Note,
    ParamCurve,
    Params,
    Phones,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
    Track,
)
from libresvip.model.point import Point
from libresvip.utils.music_math import (
    clamp,
    cosine_easing_in_out_interpolation,
    cubic_interpolation,
    db_to_float,
    linear_interpolation,
    ratio_to_db,
)
from libresvip.utils.search import find_index

from .constants import TICK_RATE
from .interval_utils import position_to_ticks
from .model import (
    SVDatabase,
    SVGroup,
    SVMeter,
    SVNote,
    SVParamCurve,
    SVParameters,
    SVPitchControl,
    SVProject,
    SVTempo,
    SVTrack,
    SVVoice,
)
from .options import BreathOption, GroupOption, InputOptions, PitchOption
from .param_expression import CurveGenerator, ParamExpression, PitchGenerator
from .phoneme_utils import default_phone_marks, sv_g2p, xsampa2pinyin
from .track_merge_utils import track_override_with

clip = cast("Callable[[int], int]", partial(clamp, lower=-1000, upper=1000))


@dataclasses.dataclass
class SynthVParser:
    options: InputOptions
    first_bar_tick: int = dataclasses.field(init=False)
    first_bpm: float = dataclasses.field(init=False)
    synchronizer: TimeSynchronizer = dataclasses.field(init=False)
    voice_settings: SVVoice = dataclasses.field(init=False)
    instant_pitch: SVParamCurve = dataclasses.field(init=False)
    group_library: dict[str, SVGroup] = dataclasses.field(default_factory=dict)
    group_split_counts: dict[str, int] = dataclasses.field(default_factory=dict)
    tracks_from_groups: list[Track] = dataclasses.field(default_factory=list)

    def actual_value_at(
        self,
        compound_expr: ParamExpression,
        mapping_func: Callable[[float], int],
        ticks: int,
    ) -> int:
        return clip(mapping_func(compound_expr.value_at_ticks(ticks) / 1000))

    @staticmethod
    def parse_interpolation(
        mode: str,
    ) -> Callable[[float, tuple[float, float], tuple[float, float]], float]:
        if mode == "cosine":
            return cosine_easing_in_out_interpolation
        elif mode == "cubic":
            return cubic_interpolation
        else:
            return linear_interpolation

    def parse_audio_offset(self, offset: int) -> int:
        if offset >= 0:
            return position_to_ticks(offset)
        return round(offset / TICK_RATE * self.first_bpm / DEFAULT_BPM)

    @staticmethod
    def parse_meter(meter: SVMeter) -> TimeSignature:
        return TimeSignature(
            bar_index=meter.index,
            numerator=meter.numerator,
            denominator=meter.denominator,
        )

    @staticmethod
    def parse_tempo(tempo: SVTempo) -> SongTempo:
        return SongTempo(position=position_to_ticks(tempo.position), bpm=tempo.bpm)

    @staticmethod
    def parse_volume(gain: float) -> float:
        if gain >= 0:
            return min(gain / (ratio_to_db(4)) + 1.0, 2.0)
        else:
            return db_to_float(gain)

    def parse_param_curve(
        self,
        sv_curve: SVParamCurve,
        mapping_func: Callable[[float], int],
        base_value: float | None = None,
        master_curve: SVParamCurve | None = None,
    ) -> ParamCurve:
        if base_value is None:
            base_value = 0.0
        curve = ParamCurve()
        interpolation_func = self.parse_interpolation(sv_curve.mode)
        decoded_base_value = mapping_func(base_value)

        generator = CurveGenerator(
            _point_list=[
                Point(
                    position_to_ticks(point.offset) + self.first_bar_tick,
                    mapping_func(point.value + base_value),
                )
                for point in sv_curve.points.root
            ],
            _interpolation=interpolation_func,
            _base_value=decoded_base_value,
        )
        if master_curve is None or not len(master_curve.points):
            curve.points.root = [
                Point(point.x, clip(point.y)) for point in generator.get_converted_curve(5)
            ]
            return curve
        if not len(sv_curve.points):
            master_generator = CurveGenerator(
                _point_list=[
                    Point(
                        position_to_ticks(point.offset) + self.first_bar_tick,
                        mapping_func(point.value + base_value),
                    )
                    for point in master_curve.points.root
                ],
                _interpolation=interpolation_func,
                _base_value=decoded_base_value,
            )
            curve.points.root = [
                Point(point.x, clip(point.y)) for point in master_generator.get_converted_curve(5)
            ]
            return curve
        group_expr = CurveGenerator(
            _point_list=[
                Point(
                    position_to_ticks(point.offset) + self.first_bar_tick,
                    round(point.value * 1000),
                )
                for point in sv_curve.points.root
            ],
            _interpolation=self.parse_interpolation(sv_curve.mode),
            _base_value=round(base_value * 1000),
        )
        master_expr = CurveGenerator(
            _point_list=[
                Point(
                    position_to_ticks(point.offset) + self.first_bar_tick,
                    round(point.value * 1000),
                )
                for point in master_curve.points.root
            ],
            _interpolation=self.parse_interpolation(master_curve.mode),
            _base_value=0,
        )
        compound_expr = group_expr + master_expr
        group_points = group_expr.point_list
        master_points = master_expr.point_list

        i = j = 0
        if group_points[i].x < master_points[j].x:
            prev_point = group_points[i]
            i += 1
            prev_point_is_base = prev_point.y == round(base_value * 1000)
        else:
            prev_point = master_points[j]
            j += 1
            prev_point_is_base = prev_point.y == 0
        curve.points.append(Point.start_point(self.actual_value_at(compound_expr, mapping_func, 0)))
        curve.points.append(
            Point(
                prev_point.x,
                self.actual_value_at(compound_expr, mapping_func, prev_point.x),
            )
        )

        while i < len(group_points) or j < len(master_points):
            if i < len(group_points) and (
                j >= len(master_points) or group_points[i].x <= master_points[j].x
            ):
                current_point = group_points[i]
                i += 1
                current_point_is_base = current_point.y == round(base_value * 1000)
            else:
                current_point = master_points[j]
                j += 1
                current_point_is_base = current_point.y == 0
            if prev_point_is_base and current_point_is_base and prev_point.x <= current_point.x:
                curve.points.append(
                    Point(
                        prev_point.x,
                        self.actual_value_at(compound_expr, mapping_func, prev_point.x),
                    )
                )
                curve.points.append(
                    Point(
                        current_point.x,
                        self.actual_value_at(compound_expr, mapping_func, current_point.x),
                    )
                )
            else:
                for p in range(prev_point.x, current_point.x, 5):
                    curve.points.append(
                        Point(
                            p,
                            self.actual_value_at(compound_expr, mapping_func, p),
                        )
                    )
            prev_point = current_point
            prev_point_is_base = current_point_is_base
        curve.points.append(
            Point(
                prev_point.x,
                self.actual_value_at(compound_expr, mapping_func, prev_point.x),
            )
        )
        curve.points.append(
            Point.end_point(
                self.actual_value_at(compound_expr, mapping_func, prev_point.x),
            )
        )
        return curve

    def parse_pitch_curve(
        self,
        pitch_diff: SVParamCurve,
        vibrato_env: SVParamCurve,
        sv_notes: list[SVNote],
        step: int = 5,
        master_pitch_diff: SVParamCurve | None = None,
        master_vibrato_env: SVParamCurve | None = None,
        pitch_controls: list[SVPitchControl] | None = None,
    ) -> ParamCurve:
        curve = ParamCurve()
        if not sv_notes:
            curve.points.append(Point.start_point())
            curve.points.append(Point.end_point())
            return curve
        pitch_diff_expr: ParamExpression = CurveGenerator(
            _point_list=[
                Point(
                    position_to_ticks(point.offset),
                    round(point.value),
                )
                for point in pitch_diff.points.root
            ],
            _interpolation=self.parse_interpolation(pitch_diff.mode),
            _base_value=0,
        )
        vibrato_env_expr: ParamExpression = CurveGenerator(
            _point_list=[
                Point(
                    position_to_ticks(point.offset),
                    round(point.value * 1000),
                )
                for point in vibrato_env.points.root
            ],
            _interpolation=self.parse_interpolation(vibrato_env.mode),
            _base_value=1000,
        )
        if master_pitch_diff is not None:
            pitch_diff_expr += CurveGenerator(
                _point_list=[
                    Point(
                        position_to_ticks(point.offset),
                        round(point.value),
                    )
                    for point in master_pitch_diff.points.root
                ],
                _interpolation=self.parse_interpolation(master_pitch_diff.mode),
                _base_value=0,
            )
        if master_vibrato_env is not None:
            vibrato_env_expr += CurveGenerator(
                _point_list=[
                    Point(
                        position_to_ticks(point.offset),
                        round(point.value * 1000),
                    )
                    for point in master_vibrato_env.points.root
                ],
                _interpolation=self.parse_interpolation(vibrato_env.mode),
                _base_value=0,
            )
        if self.options.instant:
            instant_interval = RangeInterval(
                [
                    (
                        position_to_ticks(note.onset),
                        position_to_ticks(note.onset + note.duration),
                    )
                    for note in sv_notes
                    if note.instant_mode is not False
                ]
            )
            pitch_diff_expr += CurveGenerator(
                interval=instant_interval.interval,
                _point_list=[
                    Point(
                        position_to_ticks(point.offset),
                        round(point.value),
                    )
                    for point in self.instant_pitch.points.root
                ],
                _interpolation=self.parse_interpolation(self.instant_pitch.mode),
                _base_value=0,
            )
        interval = RangeInterval(
            [
                (
                    position_to_ticks(note.onset),
                    position_to_ticks(note.onset + note.duration),
                )
                for note in sv_notes
            ]
        ).expand(120)

        generator = PitchGenerator(
            _synchronizer=self.synchronizer,
            _note_list=sv_notes,
            _pitch_diff=pitch_diff_expr,
            _vibrato_env=vibrato_env_expr,
            _pitch_controls=pitch_controls,
        )
        if self.options.pitch in {PitchOption.VIBRATO, PitchOption.PLAIN}:
            regard_default_vibrato_as_unedited = self.options.pitch == PitchOption.PLAIN
            pitch_interval = 0.1

            def reduced_interval(current: RangeInterval, note: SVNote) -> RangeInterval:
                start_secs = (
                    self.synchronizer.get_actual_secs_from_ticks(position_to_ticks(note.onset))
                    - max(0.0, note.attributes.transition_offset)
                    - pitch_interval
                )
                end_secs = (
                    self.synchronizer.get_actual_secs_from_ticks(
                        position_to_ticks(note.onset + note.duration)
                    )
                    + pitch_interval
                )
                return current | RangeInterval(
                    [
                        (
                            round(
                                self.synchronizer.get_actual_ticks_from_secs(max(0.0, start_secs))
                            ),
                            round(self.synchronizer.get_actual_ticks_from_secs(end_secs)),
                        )
                    ]
                )

            note_edited_range = reduce(
                reduced_interval,
                (
                    note
                    for note in sv_notes
                    if note.pitch_edited(
                        regard_default_vibrato_as_unedited,
                        self.options.instant,
                    )
                ),
                RangeInterval(),
            )
            param_edited_range = pitch_diff.edited_range() | vibrato_env.edited_range(1.0)
            if master_pitch_diff is not None:
                param_edited_range |= master_pitch_diff.edited_range()
            if generator.pitch_control_layer is not None:
                param_edited_range.interval |= generator.pitch_control_layer.interval_dict.domain()
            if master_vibrato_env is not None:
                param_edited_range |= master_vibrato_env.edited_range(1.0)
            interval &= note_edited_range | param_edited_range
        curve.points.append(Point.start_point())
        for start, end in interval.shift(self.first_bar_tick).sub_ranges():
            curve.points.append(Point(start, -100))
            curve.points.extend(
                Point(i, round(generator.value_at_ticks(i - self.first_bar_tick)))
                for i in range(start, end, step)
            )
            curve.points.append(
                Point(
                    end,
                    round(generator.value_at_ticks(end - self.first_bar_tick)),
                )
            )
            curve.points.append(Point(end, -100))
        curve.points.append(Point.end_point())
        return curve

    def parse_params(
        self,
        sv_params: SVParameters,
        sv_notes: list[SVNote],
        master_params: SVParameters | None = None,
        pitch_controls: list[SVPitchControl] | None = None,
    ) -> Params:
        params = Params()
        if self.options.import_pitch:
            params.pitch = self.parse_pitch_curve(
                sv_params.pitch_delta,
                sv_params.vibrato_env,
                sv_notes,
                5,
                master_params.pitch_delta if master_params else None,
                master_params.vibrato_env if master_params else None,
                pitch_controls or None,
            )
        if self.options.import_volume:
            params.volume = self.parse_param_curve(
                sv_params.loudness,
                lambda val: round(val / 12.0 * 1000.0)
                if val >= 0.0
                else round(1000 * db_to_float(val) - 1000),
                self.voice_settings.param_loudness or 0.0,
                master_params.loudness if master_params else None,
            )
        if self.options.import_breath:
            params.breath = self.parse_param_curve(
                sv_params.breathiness,
                lambda val: round(val * 1000),
                self.voice_settings.param_breathiness or 0.0,
                master_params.breathiness if master_params else None,
            )
        if self.options.import_gender:
            params.gender = self.parse_param_curve(
                sv_params.gender,
                lambda val: round(val * -1000),
                self.voice_settings.param_gender or 0.0,
                master_params.gender if master_params else None,
            )
        if self.options.import_strength:
            params.strength = self.parse_param_curve(
                sv_params.tension,
                lambda val: round(val * 1000),
                self.voice_settings.param_tension or 0.0,
                master_params.tension if master_params else None,
            )
        return params

    @staticmethod
    def parse_note(sv_note: SVNote, database: SVDatabase) -> Note:
        note = Note(
            start_pos=position_to_ticks(sv_note.onset),
            key_number=sv_note.pitch,
        )
        note.length = position_to_ticks(sv_note.onset + sv_note.duration) - note.start_pos
        note.lyric = SVNote.normalize_lyric(sv_note.lyrics)
        if sv_note.phonemes:
            note_default_language = sv_note.attributes.default_language(database)
            if note_default_language in ["japanese", "korean"]:
                note.pronunciation = sv_note.phonemes
            elif note_default_language in ["mandarin", "cantonese"]:
                note.pronunciation = xsampa2pinyin(sv_note.phonemes, note_default_language)
        return note

    def parse_note_list(self, sv_note_list: list[SVNote], database: SVDatabase) -> list[Note]:
        note_list = []
        breath_pattern = re.compile(r"^\s*\.?\s*br(l?[1-9])?\s*$")
        if self.options.breath == BreathOption.CONVERT:
            if len(sv_note_list) > 1:
                prev_index = -1
                while (
                    current_offset := find_index(
                        sv_note_list[prev_index + 1 :],
                        lambda _note: breath_pattern.match(_note.lyrics) is None,
                    )
                ) != -1:
                    current_index = prev_index + current_offset + 1
                    note = self.parse_note(sv_note_list[current_index], database)
                    if current_index > prev_index + 1:
                        breath_note = sv_note_list[current_index - 1]
                        if (
                            position_to_ticks(sv_note_list[current_index].onset)
                            - position_to_ticks(breath_note.onset + breath_note.duration)
                        ) < 120:
                            note.head_tag = "V"
                    note_list.append(note)
                    prev_index = current_index
            elif len(sv_note_list) == 1 and breath_pattern.match(sv_note_list[0].lyrics) is None:
                note_list.append(self.parse_note(sv_note_list[0], database))
            sv_note_list = [
                note for note in sv_note_list if breath_pattern.match(note.lyrics) is None
            ]
        else:
            if self.options.breath == BreathOption.IGNORE:
                sv_note_list = [
                    note for note in sv_note_list if breath_pattern.match(note.lyrics) is None
                ]
            note_list = [self.parse_note(note, database) for note in sv_note_list]
        if len(sv_note_list):
            lyrics, languages = zip(
                *(
                    (
                        SVNote.normalize_phoneme(note),
                        sv_note.attributes.default_language(database),
                    )
                    for note, sv_note in zip(note_list, sv_note_list)
                )
            )
        else:
            lyrics, languages = (), ()
        lyrics_phoneme = sv_g2p(lyrics, languages)
        if not len(note_list):
            return note_list
        current_sv_note = sv_note_list[0]
        current_duration = current_sv_note.attributes.dur
        current_phonemes = current_sv_note.attributes.phonemes
        current_phone_marks = default_phone_marks(
            lyrics_phoneme[0],
            current_sv_note.attributes.default_language(database),
        )

        if (
            current_phonemes is not None
            and len(current_phonemes) > 0
            and current_phonemes[0].left_offset is not None
        ):
            note_list[0].edited_phones = Phones(
                head_length_in_secs=min(1.8, -current_phonemes[0].left_offset),
            )
        elif (
            current_phone_marks[0] > 0
            and current_duration is not None
            and current_duration[0] != 1.0
        ):
            note_list[0].edited_phones = Phones(
                head_length_in_secs=min(1.8, current_duration[0] * current_phone_marks[0]),
            )

        for i in range(len(sv_note_list) - 1):
            next_sv_note = sv_note_list[i + 1]
            next_duration = next_sv_note.attributes.dur
            next_phonemes = next_sv_note.attributes.phonemes
            next_phone_marks = default_phone_marks(
                lyrics_phoneme[i + 1],
                next_sv_note.attributes.default_language(database),
            )

            index = 1 if next_phone_marks[0] > 0 else 0
            if (
                next_phonemes is not None
                and len(next_phonemes)
                and next_phonemes[0].left_offset is not None
            ):
                next_head_part_length = -next_phonemes[0].left_offset
            elif next_phone_marks[0] > 0 and next_duration is not None and len(next_duration):
                next_head_part_length = next_duration[0] * next_phone_marks[0]
            else:
                next_head_part_length = None
            if (
                current_phone_marks[1] > 0
                and current_duration is not None
                and len(current_duration) > index + 1
            ):
                if note_list[i].edited_phones is None:
                    note_list[i].edited_phones = Phones()
                note_list[i].edited_phones.mid_ratio_over_tail = (  #  type: ignore[union-attr]
                    current_phone_marks[1] * current_duration[index] / current_duration[index + 1]
                )
            if next_head_part_length is not None:
                if note_list[i + 1].edited_phones is None:
                    note_list[i + 1].edited_phones = Phones()
                space_in_secs = min(
                    2.0,
                    self.synchronizer.get_duration_secs_from_ticks(
                        note_list[i].start_pos + self.first_bar_tick,
                        note_list[i + 1].start_pos + self.first_bar_tick,
                    ),
                )
                if (
                    current_phone_marks[1] > 0
                    and current_duration is not None
                    and len(current_duration) > index
                ):
                    ratio = (
                        2 / (current_duration[index] + current_duration[index + 1])
                        if len(current_duration) > index + 1
                        else 1 / current_duration[index]
                    )
                    if (
                        next_head_part_length * ratio
                        > self.synchronizer.get_duration_secs_from_ticks(
                            note_list[i].end_pos + self.first_bar_tick,
                            note_list[i + 1].start_pos + self.first_bar_tick,
                        )
                    ):
                        next_head_part_length *= ratio
                note_list[i + 1].edited_phones.head_length_in_secs = min(  #  type: ignore[union-attr]
                    0.9 * space_in_secs, next_head_part_length
                )
            current_duration = next_duration
            current_phone_marks = next_phone_marks
        index = 1 if current_phone_marks[0] > 0 else 0
        if (
            current_phone_marks[1] > 0
            and current_duration is not None
            and len(current_duration) > index + 1
        ):
            if note_list[-1].edited_phones is None:
                note_list[-1].edited_phones = Phones()
            note_list[-1].edited_phones.mid_ratio_over_tail = (
                current_phone_marks[1] * current_duration[index] / current_duration[index + 1]
            )
        return note_list

    def parse_track(self, sv_track: SVTrack) -> Track | None:
        if sv_track.main_ref.is_instrumental:
            if self.options.import_instrumental_track and sv_track.main_ref.audio is not None:
                svip_track = InstrumentalTrack(
                    audio_file_path=sv_track.main_ref.audio.filename,
                    offset=self.parse_audio_offset(sv_track.main_ref.blick_offset),
                )
            else:
                return None
        else:
            self.voice_settings = sv_track.main_ref.voice
            master_note_attributes = self.voice_settings.to_attributes()
            if self.options.instant:
                self.instant_pitch = sv_track.main_ref.system_pitch_delta
            for note in sv_track.main_group.notes:
                note.merge_attributes(master_note_attributes)
            singing_track = SingingTrack(
                ai_singer_name=sv_track.main_ref.database.name,
                note_list=self.parse_note_list(
                    sv_track.main_group.notes, sv_track.main_ref.database
                ),
                edited_params=self.parse_params(
                    sv_track.main_group.parameters, sv_track.main_group.notes
                ),
            )
            if self.options.group == GroupOption.SPLIT:
                for sv_ref in sv_track.groups:
                    group = (
                        self.group_library[sv_ref.group_id] + sv_ref.blick_offset
                        ^ sv_ref.pitch_offset
                    )
                    voice_settings = sv_ref.voice
                    master_note_attributes = voice_settings.to_attributes()
                    if self.options.instant:
                        self.instant_pitch = sv_ref.system_pitch_delta
                    for note in group.notes:
                        note.merge_attributes(master_note_attributes)
                    self.group_split_counts[sv_ref.group_id] += 1
                    self.tracks_from_groups.append(
                        SingingTrack(
                            ai_singer_name=sv_ref.database.name,
                            title=f"{group.name} ({self.group_split_counts[sv_ref.group_id]})",
                            note_list=self.parse_note_list(group.notes, sv_ref.database),
                            edited_params=self.parse_params(
                                group.parameters,
                                group.notes,
                                sv_track.main_group.parameters,
                                group.pitch_controls,
                            ),
                        )
                    )
            elif self.options.group == GroupOption.MERGE:
                merged_group = sv_track.main_group
                for sv_ref in sv_track.groups:
                    group = (
                        self.group_library[sv_ref.group_id] + sv_ref.blick_offset
                        ^ sv_ref.pitch_offset
                    )
                    voice_settings = sv_ref.voice
                    master_note_attributes = voice_settings.to_attributes()
                    if self.options.instant:
                        self.instant_pitch = sv_ref.system_pitch_delta
                    for note in group.notes:
                        note.merge_attributes(master_note_attributes)
                    if merged_group.overlapped_with(group):
                        self.group_split_counts[sv_ref.group_id] += 1
                        self.tracks_from_groups.append(
                            SingingTrack(
                                ai_singer_name=sv_ref.database.name,
                                title=f"{group.name} ({self.group_split_counts[sv_ref.group_id]})",
                                note_list=self.parse_note_list(group.notes, sv_ref.database),
                                edited_params=self.parse_params(
                                    group.parameters,
                                    group.notes,
                                    sv_track.main_group.parameters,
                                    group.pitch_controls,
                                ),
                            )
                        )
                    else:
                        track_override_with(
                            singing_track,
                            self.parse_note_list(group.notes, sv_ref.database),
                            self.parse_params(
                                group.parameters,
                                group.notes,
                                sv_track.main_group.parameters,
                                group.pitch_controls,
                            ),
                            self.first_bar_tick,
                        )
                        merged_group.notes.extend(group.notes)
                        merged_group.notes.sort(key=operator.attrgetter("onset"))
            svip_track = singing_track
        svip_track.title = sv_track.name
        svip_track.mute = sv_track.mixer.mute
        svip_track.solo = sv_track.mixer.solo
        svip_track.pan = sv_track.mixer.pan
        svip_track.volume = self.parse_volume(sv_track.mixer.gain_decibel)
        return svip_track

    def parse_project(self, sv_project: SVProject) -> Project:
        project = Project()
        time_sig = sv_project.time_sig
        self.first_bpm = time_sig.tempo[0].bpm

        project.time_signature_list = shift_beat_list(
            [self.parse_meter(meter) for meter in time_sig.meter], 1
        )
        self.first_bar_tick = round(project.time_signature_list[0].bar_length())
        project.song_tempo_list = shift_tempo_list(
            [self.parse_tempo(tempo) for tempo in time_sig.tempo],
            self.first_bar_tick,
        )
        self.synchronizer = TimeSynchronizer(project.song_tempo_list)

        for sv_group in sv_project.library:
            self.group_library[sv_group.uuid] = sv_group
            self.group_split_counts[sv_group.uuid] = 0

        for sv_track in sv_project.tracks:
            if track := self.parse_track(sv_track):
                project.track_list.append(track)

        project.track_list.extend(self.tracks_from_groups)
        return project



from __future__ import annotations

import base64
import pathlib
from typing import Any

from PySide6.QtCore import Property, QObject, Signal, Slot
from PySide6.QtQml import QmlElement

from __feature__ import snake_case, true_property  # isort:skip # noqa: F401

import libresvip
from libresvip.core.config import (
    ConflictPolicy,
    DarkMode,
    LibreSvipSettings,
    save_settings,
    settings,
)
from libresvip.core.constants import res_dir
from libresvip.extension.manager import plugin_manager
from libresvip.gui.models.list_models import (
    LyricReplacementPresetsModel,
    ModelProxy,
)
from libresvip.gui.models.table_models import LyricReplacementRulesTableModel

from .application import app

QML_IMPORT_NAME = "LibreSVIP"
QML_IMPORT_MAJOR_VERSION = 1
QML_IMPORT_MINOR_VERSION = 0


def base_prop_factory(attrs: dict[str, Any], field_name: str, field_type: type) -> None:
    signal = Signal(field_type, name=f"{field_name}_changed")

    def _getter(self: QObject) -> Any:
        return getattr(settings, field_name)

    def _setter(self: QObject, value: Any) -> None:
        setattr(settings, field_name, value)
        getattr(self, f"{field_name}_changed").emit(value)

    attrs[field_name] = Property(field_type, _getter, _setter, notify=signal)
    attrs[f"{field_name}_changed"] = signal


class AutoBindBaseConfigMetaObject(type(QObject)):  # type: ignore[misc]
    def __new__(cls, name: str, bases: tuple[type], attrs: dict[str, Any]) -> type[QObject]:
        for field_name, field_info in LibreSvipSettings.model_fields.items():
            if field_info.annotation is bool:
                base_prop_factory(attrs, field_name, bool)
            elif field_info.annotation is int:
                base_prop_factory(attrs, field_name, int)
        return super().__new__(cls, name, bases, attrs)


@QmlElement
class ConfigItems(QObject, metaclass=AutoBindBaseConfigMetaObject):
    save_folder_changed = Signal(str)
    conflict_policy_changed = Signal(str)
    theme_changed = Signal(str)

    def __init__(self, parent: QObject | None = None) -> None:
        super().__init__(parent=parent)
        self.folder_presets = ModelProxy({"path": ""})
        self.folder_presets.append_many(
            [{"path": self.posix_path(path)} for path in settings.folder_presets]
        )
        self.lyric_replacement_presets = LyricReplacementPresetsModel()
        app.aboutToQuit.connect(self.save_settings)

    def save_settings(self) -> None:
        settings.folder_presets = [pathlib.Path(item["path"]) for item in self.folder_presets.items]
        save_settings()

    @Property(str, constant=True)
    def icon_data(self) -> str:
        return f"data:image/x-icon;base64,{base64.b64encode((res_dir / 'libresvip.ico').read_bytes()).decode()}"

    @Property(str, constant=True)
    def version(self) -> str:
        return libresvip.__version__

    @Slot(str, result="QVariant")
    def qget(self, name: str) -> Any:
        return getattr(self, name)

    @Slot(str, result="QVariant")
    def rules_for_preset(self, preset: str) -> Any:
        self.result = LyricReplacementRulesTableModel(preset)
        return self.result

    def get_save_folder(self) -> str:
        return self.posix_path(settings.save_folder)

    def set_save_folder(self, value: str) -> None:
        settings.save_folder = pathlib.Path(value)
        self.save_folder_changed.emit(self.get_save_folder())

    save_folder = Property(str, get_save_folder, set_save_folder, notify=save_folder_changed)

    def get_conflict_policy(self) -> str:
        return settings.conflict_policy.value

    def set_conflict_policy(self, policy: str) -> None:
        conflict_policy = ConflictPolicy(policy)
        settings.conflict_policy = conflict_policy
        self.conflict_policy_changed.emit(policy)

    conflict_policy = Property(
        str,
        get_conflict_policy,
        set_conflict_policy,
        notify=conflict_policy_changed,
    )

    def get_theme(self) -> str:
        return settings.dark_mode.value

    def set_theme(self, theme: str) -> None:
        dark_mode = DarkMode(theme)
        settings.dark_mode = dark_mode
        self.theme_changed.emit(theme)

    theme = Property(str, get_theme, set_theme, notify=theme_changed)

    @Slot(str, result=bool)
    def enabled(self, key: str) -> bool:
        return key in plugin_manager.plugin_registry

    @staticmethod
    def posix_path(path: pathlib.Path) -> str:
        return str(path.as_posix())

    @Slot(str, result=bool)
    def dir_valid(self, value: str) -> bool:
        path = pathlib.Path(value)
        return (not path.is_absolute() or path.exists()) and path.is_dir()



from __future__ import annotations

from dataclasses import dataclass
from functools import partial
from typing import TYPE_CHECKING

from libresvip.utils.music_math import (
    cosine_easing_in_out_interpolation,
    cubic_interpolation,
    linear_interpolation,
    sigmoid_interpolation,
    vocaloid_interpolation,
)

if TYPE_CHECKING:
    from collections.abc import Callable


@dataclass
class PortamentoPitch:
    max_inter_time_in_secs: float
    max_inter_time_percent: float
    inter_func: Callable[[float, tuple[float, float], tuple[float, float]], float]
    vocaloid_mode: bool = False

    @classmethod
    def no_portamento(cls) -> PortamentoPitch:
        return cls(0.0, 0.0, linear_interpolation)

    @classmethod
    def cosine_portamento(cls) -> PortamentoPitch:
        return cls(0.05, 0.1, cosine_easing_in_out_interpolation)

    @classmethod
    def cubic_portamento(cls) -> PortamentoPitch:
        return cls(0.05, 0.1, cubic_interpolation)

    @classmethod
    def sigmoid_portamento(cls) -> PortamentoPitch:
        return cls(0.075, 0.48, partial(sigmoid_interpolation, k=5.5))  # type: ignore[call-arg]

    @classmethod
    def vocaloid_portamento(cls) -> PortamentoPitch:
        return cls(0.05, 0.15, vocaloid_interpolation, True)



from pydantic import BaseModel, Field

from libresvip.model.option_mixins import (
    EnablePitchImportationMixin,
    SelectSingleTrackMixin,
    StaticTempoMixin,
)
from libresvip.utils.translation import gettext_lazy as _


class InputOptions(EnablePitchImportationMixin, BaseModel):
    pass


class OutputOptions(SelectSingleTrackMixin, StaticTempoMixin, BaseModel):
    version: int = Field(
        default=19,
        title=_("Version"),
        description=_("Version of NIAONiao project file"),
    )



from __future__ import annotations

import dataclasses
import operator
from functools import singledispatch
from itertools import groupby
from typing import SupportsFloat

from libresvip.core.time_sync import TimeSynchronizer
from libresvip.model.base import Note, ParamCurve, SongTempo
from libresvip.model.point import Point
from libresvip.utils.music_math import (
    interpolate_cosine_ease_in_out,
    interpolate_linear,
)

from .constants import (
    BEND_DOWN_LENGTH_FIXED_SEC,
    BEND_LENGTH_MAX_SEC,
    BEND_LENGTH_MIN_SEC,
    BEND_VALUE_MAX,
    NOTE_KEY_SUM,
    PORTAMENTO_LENGTH_MAX_SEC,
    SAMPLING_INTERVAL_TICK,
)
from .model import DvPoint


@dataclasses.dataclass
class DvSegmentPitchRawData:
    tick_offset: int  # only for import
    data: list[DvPoint]  # (tick, DV style cent)


@dataclasses.dataclass
class DvNoteWithPitch:
    note: Note
    por_head: int  # 0~100
    por_tail: int  # 0~100
    ben_len: int  # 0~100
    ben_dep: int  # 0~100
    vibrato: list[DvPoint]  # (ms, minus cent)


def merge_points_from_segments(
    segments: list[DvSegmentPitchRawData],
) -> list[Point] | None:
    points = []
    for segment in segments:
        for dv_point in segment.data:
            if (raw_tick := dv_point.x) >= 0:
                tick = raw_tick + segment.tick_offset
                if dv_point.y >= 0:
                    value = round(convert_note_key(dv_point.y / 100) * 100)
                    points.append(Point(x=tick, y=value))
                else:
                    points.append(Point(x=tick, y=-100))
    return points or None


def merge_same_tick_points(points: list[Point]) -> list[Point] | None:
    merged_points = []
    for tick, group in groupby(iter(points), key=operator.attrgetter("x")):
        group_list = list(group)
        if len(group_list) > 1:
            if any(point.y == -100 for point in group_list):
                merged_points.append(Point(x=tick, y=-100))
            else:
                merged_points.append(
                    Point(
                        x=tick,
                        y=round(sum(point.y for point in group_list) / len(group_list)),
                    )
                )
        else:
            merged_points.append(group_list[0])
    return merged_points or None


def merge_same_value_points(points: list[Point]) -> list[Point] | None:
    merged_points = [next(group) for key, group in groupby(points, key=operator.attrgetter("y"))]
    return merged_points or None


def apply_default_pitch(
    first_bar_length: int,
    points: list[Point],
    notes: list[DvNoteWithPitch],
    tempos: list[SongTempo],
) -> list[Point]:
    if not points or not notes:
        return points

    time_synchronizer = TimeSynchronizer(tempos)

    base = get_base_pitch(notes, time_synchronizer)
    bend_diff = get_bend_pitch(notes, time_synchronizer)
    vibrato_diff = get_vibrato_pitch(notes, time_synchronizer)

    result = []
    last_point = None
    for point in points:
        if last_point is not None and last_point.y == -100:
            result.extend(
                Point(
                    x=tick + first_bar_length,
                    y=base.get(tick, 0) + bend_diff.get(tick, 0) + vibrato_diff.get(tick, 0),
                )
                for tick in range(last_point.x + 1, point.x, SAMPLING_INTERVAL_TICK)
            )
            result.append(
                Point(
                    x=(point.x if last_point is None else last_point.x) + first_bar_length,
                    y=-100,
                )
            )
        if point.y != -100:
            result.append(point._replace(x=point.x + first_bar_length))
        else:
            result.append(Point(x=point.x + first_bar_length, y=-100))
        last_point = point
    if not result or result[0].x > notes[-1].note.end_pos:
        result = [
            Point(
                x=tick + first_bar_length,
                y=(base.get(tick, 0) + bend_diff.get(tick, 0) + vibrato_diff.get(tick, 0)),
            )
            for tick in range(0, notes[-1].note.end_pos, SAMPLING_INTERVAL_TICK)
        ] + result
    if result:
        result.insert(0, Point.start_point())
        result.append(Point.end_point())
    return result


def get_base_pitch(notes: list[DvNoteWithPitch], transformer: TimeSynchronizer) -> dict[int, int]:
    result = {}
    for last_note, this_note in zip([None, *notes], [*notes, None]):
        portamento = []
        if last_note is not None and this_note is not None:
            portamento = get_portamento(last_note, transformer, this_note)
            for point in portamento:
                result[point.x] = point.y
        if last_note is not None:
            last_note_tail = [
                (tick, last_note.note.key_number * 100)
                for tick in range(
                    tick_half_start(last_note.note),
                    (portamento[0].x if portamento else last_note.note.end_pos),
                )
            ]
            result |= last_note_tail
        if this_note is not None:
            start = (
                0
                if last_note is None
                else (portamento[-1].x if portamento else this_note.note.start_pos)
            )
            this_note_head = [
                (tick, this_note.note.key_number * 100)
                for tick in range(start, tick_half_start(this_note.note))
            ]
            result.update(this_note_head)
    return result


def get_bend_pitch(notes: list[DvNoteWithPitch], transformer: TimeSynchronizer) -> dict[int, int]:
    result: dict[int, int] = {}
    for note in notes:
        start_tick = note.note.start_pos
        start_sec = transformer.get_actual_secs_from_ticks(start_tick)
        valley_sec = start_sec + BEND_DOWN_LENGTH_FIXED_SEC
        valley_tick = min(
            transformer.get_actual_ticks_from_secs(valley_sec),
            note.note.start_pos + note.note.length // 2 - 1,
        )

        if note.ben_len <= 50:
            length_sec = BEND_LENGTH_MIN_SEC
        else:
            length_sec = (BEND_LENGTH_MAX_SEC - BEND_LENGTH_MIN_SEC) * (
                note.ben_len - 50
            ) // 50 + BEND_LENGTH_MIN_SEC
        end_sec = start_sec + length_sec
        end_tick = min(
            transformer.get_actual_ticks_from_secs(end_sec),
            note.note.start_pos - 1,
        )

        valley_value = -BEND_VALUE_MAX * note.ben_dep
        valley_point = Point(x=round(valley_tick), y=round(valley_value))

        bend_down = interpolate_linear([Point(x=start_tick, y=0), valley_point], 1) or []

        bend_up = (
            interpolate_cosine_ease_in_out([valley_point, Point(x=round(end_tick), y=0)], 1)[1:]
            or []
        )

        result |= dict(bend_down + bend_up)
    return result


def get_portamento(
    last_note: DvNoteWithPitch,
    transformer: TimeSynchronizer,
    this_note: DvNoteWithPitch,
) -> list[Point]:
    tail_length_sec = PORTAMENTO_LENGTH_MAX_SEC * last_note.por_tail // 100
    start_sec = transformer.get_actual_secs_from_ticks(last_note.note.end_pos) - tail_length_sec
    start_tick = max(
        transformer.get_actual_ticks_from_secs(start_sec),
        tick_half_start(last_note.note),
    )

    head_length_sec = PORTAMENTO_LENGTH_MAX_SEC * this_note.por_head // 100
    end_sec = transformer.get_actual_secs_from_ticks(this_note.note.start_pos) + head_length_sec
    end_tick = min(
        transformer.get_actual_ticks_from_secs(end_sec),
        tick_half_start(this_note.note) - 1,
    )

    return (
        interpolate_cosine_ease_in_out(
            [
                Point(round(start_tick), last_note.note.key_number * 100),
                Point(round(end_tick), this_note.note.key_number * 100),
            ],
            1,
        )
        or []
    )


def get_vibrato_pitch(
    notes: list[DvNoteWithPitch], transformer: TimeSynchronizer
) -> dict[int, int]:
    result: dict[int, int] = {}
    for note in notes:
        start_tick = note.note.start_pos
        start_sec = transformer.get_actual_secs_from_ticks(start_tick)
        vibrato_points = [
            Point(
                x=round(transformer.get_actual_ticks_from_secs(start_sec + vib_point.x / 1000)),
                y=-vib_point.y,
            )
            for vib_point in note.vibrato
            if start_tick
            <= transformer.get_actual_ticks_from_secs(start_sec + vib_point.x / 1000)
            < note.note.end_pos
        ]
        result |= dict(interpolate_linear(vibrato_points, 1) or [])
    return result


def tick_half_start(note: Note) -> int:
    return (note.start_pos + note.end_pos) // 2


@singledispatch
def convert_note_key(key: SupportsFloat) -> float:
    raise NotImplementedError


@convert_note_key.register(int)
def _convert_note_key_int(key: int) -> int:
    return int(NOTE_KEY_SUM) - key


@convert_note_key.register(float)
def _convert_note_key_float(key: float) -> float:
    return NOTE_KEY_SUM - key


def pitch_from_dv_track(
    first_bar_length: int,
    segments: list[DvSegmentPitchRawData],
    notes: list[DvNoteWithPitch],
    tempos: list[SongTempo],
) -> ParamCurve | None:
    merged_points = merge_points_from_segments(segments)
    if merged_points is None:
        return None
    merged_points = merge_same_tick_points(merged_points)
    if merged_points is None:
        return None
    merged_points = merge_same_value_points(merged_points)
    if merged_points is None:
        return None
    return ParamCurve(points=apply_default_pitch(first_bar_length, merged_points, notes, tempos))


def generate_for_dv(
    first_bar_length: int, pitch: ParamCurve, notes: list[Note]
) -> DvSegmentPitchRawData | None:
    if not len(pitch.points):
        return None
    data = [DvPoint(x=-1, y=-1)]
    last_value = None
    for point in pitch.points.root:
        if (last_value is None and point.y != -100) or point.y == -100:
            data.append(DvPoint(x=point.x - first_bar_length, y=-1))
        if point.y != -100:
            data.append(
                DvPoint(
                    x=point.x - first_bar_length,
                    y=round(convert_note_key(point.y / 100) * 100),
                )
            )
        last_value = point.y
    data.append(DvPoint(x=data[-1].x + 1 if len(data) > 1 else 307201, y=-1))
    return DvSegmentPitchRawData(0, data)



from __future__ import annotations

import abc
import dataclasses
import enum
import operator
from typing import TYPE_CHECKING

from libresvip.model.point import Point
from libresvip.utils.search import binary_find_last

from .interval_utils import position_to_ticks
from .layer_generator import (
    BaseLayerGenerator,
    GaussianLayerGenerator,
    NoteStruct,
    PitchControlLayerGenerator,
    VibratoLayerGenerator,
)

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable

    import portion

    from libresvip.core.time_sync import TimeSynchronizer

    from .model import SVNote, SVPitchControl


class ParamOperators(enum.Enum):
    ADD = operator.add
    SUB = operator.sub
    MUL = operator.mul
    DIV = operator.truediv


class ParamExpression(abc.ABC):
    @abc.abstractmethod
    def value_at_ticks(self, ticks: int) -> float:
        pass

    def __add__(self, other: int | ParamExpression) -> ParamExpression:
        if isinstance(other, int):
            return TranslationalParam(
                self,
                other,
            )
        else:
            return CompoundParam(
                self,
                ParamOperators.ADD,
                other,
            )

    def __sub__(self, other: int | ParamExpression) -> ParamExpression:
        if isinstance(other, int):
            return self + (-other)
        else:
            return CompoundParam(
                self,
                ParamOperators.SUB,
                other,
            )

    def __mul__(self, other: float | ParamExpression) -> ParamExpression:
        if isinstance(other, float):
            return ScaledParam(
                self,
                other,
            )
        else:
            return CompoundParam(
                self,
                ParamOperators.MUL,
                other,
            )

    def __truediv__(self, other: float | ParamExpression) -> ParamExpression:
        if isinstance(other, float):
            return self * (1 / other)
        else:
            return CompoundParam(
                self,
                ParamOperators.DIV,
                other,
            )


@dataclasses.dataclass
class ScaledParam(ParamExpression):
    expression: ParamExpression
    ratio: float

    def value_at_ticks(self, ticks: int) -> float:
        return self.ratio * self.expression.value_at_ticks(ticks)


@dataclasses.dataclass
class TranslationalParam(ParamExpression):
    expression: ParamExpression
    offset: int

    def value_at_ticks(self, ticks: int) -> float:
        return self.expression.value_at_ticks(ticks) + self.offset


@dataclasses.dataclass
class CurveGenerator(ParamExpression):
    base_value: int = dataclasses.field(init=False)
    interpolation: Callable[[float, tuple[float, float], tuple[float, float]], float] = (
        dataclasses.field(init=False)
    )
    point_list: list[Point] = dataclasses.field(init=False)
    pos_indexes: list[tuple[int, int]] = dataclasses.field(init=False)
    _point_list: dataclasses.InitVar[Iterable[Point]]
    _interpolation: dataclasses.InitVar[
        Callable[[float, tuple[float, float], tuple[float, float]], float]
    ]
    _base_value: dataclasses.InitVar[int]
    interval: portion.Interval | None = None

    def __post_init__(
        self,
        _point_list: Iterable[Point],
        _interpolation: Callable[[float, tuple[float, float], tuple[float, float]], float],
        _base_value: int,
    ) -> None:
        self.point_list = []
        current_pos = -1
        current_sum = 0
        overlap_count = 0
        for pos, value in _point_list:
            if pos != current_pos and current_pos >= 0:
                self.point_list.append(Point(current_pos, round(current_sum / overlap_count)))
                current_sum = 0
                overlap_count = 0
            current_sum += value
            overlap_count += 1
            current_pos = pos
        if current_pos != -1:
            self.point_list.append(Point(current_pos, round(current_sum / overlap_count)))
        self.interpolation = _interpolation
        self.base_value = _base_value
        self.pos_indexes = [(p.x, i) for i, p in enumerate(self.point_list)]

    def value_at_ticks(self, ticks: int) -> float:
        if len(self.point_list) == 0 or (self.interval is not None and ticks not in self.interval):
            return self.base_value
        index_item = binary_find_last(self.pos_indexes, lambda item: item[0] <= ticks)
        if index_item is None:
            return self.point_list[0].y
        pos, index = index_item
        if index == len(self.point_list) - 1:
            return self.point_list[-1].y
        return self.interpolation(ticks, self.point_list[index], self.point_list[index + 1])

    def get_converted_curve(self, step: int) -> list[Point]:
        result: list[Point] = []
        if len(self.point_list) == 0:
            result.extend(
                (
                    Point.start_point(self.base_value),
                    Point.end_point(self.base_value),
                )
            )
            return result
        prev_point = self.point_list[0]
        result.extend(
            (
                Point.start_point(prev_point.y),
                Point(prev_point.x, prev_point.y),
            )
        )
        for current_point in self.point_list[1:]:
            if prev_point.y == self.base_value and current_point.y == self.base_value:
                result.extend(
                    (
                        Point(prev_point.x, self.base_value),
                        Point(current_point.x, self.base_value),
                    )
                )
            else:
                for p in range(prev_point.x + step, current_point.x, step):
                    v = round(self.interpolation(p, prev_point, current_point))
                    result.append(Point(p, v))
            prev_point = current_point
        result.extend((Point(prev_point.x, prev_point.y), Point.end_point(prev_point.y)))
        return result


@dataclasses.dataclass
class CompoundParam(ParamExpression):
    expr1: ParamExpression
    op: ParamOperators
    expr2: ParamExpression

    def value_at_ticks(self, ticks: int) -> float:
        ticks1 = self.expr1.value_at_ticks(ticks)
        ticks2 = self.expr2.value_at_ticks(ticks)
        if self.op == ParamOperators.ADD:
            return ticks1 + ticks2
        elif self.op == ParamOperators.SUB:
            return ticks1 - ticks2
        elif self.op == ParamOperators.MUL:
            return ticks1 * ticks2
        elif self.op == ParamOperators.DIV:
            return ticks1 / ticks2
        raise NotImplementedError


@dataclasses.dataclass
class PitchGenerator(ParamExpression):
    synchronizer: TimeSynchronizer = dataclasses.field(init=False)
    pitch_diff: ParamExpression = dataclasses.field(init=False)
    vibrato_env: ParamExpression = dataclasses.field(init=False)
    base_layer: BaseLayerGenerator = dataclasses.field(init=False)
    gaussian_layer: GaussianLayerGenerator = dataclasses.field(init=False)
    vibrato_layer: VibratoLayerGenerator = dataclasses.field(init=False)
    _synchronizer: dataclasses.InitVar[TimeSynchronizer]
    _pitch_diff: dataclasses.InitVar[ParamExpression]
    _vibrato_env: dataclasses.InitVar[ParamExpression]
    _note_list: dataclasses.InitVar[list[SVNote]]
    _pitch_controls: dataclasses.InitVar[list[SVPitchControl] | None]
    pitch_control_layer: PitchControlLayerGenerator | None = None

    def __post_init__(
        self,
        _synchronizer: TimeSynchronizer,
        _pitch_diff: ParamExpression,
        _vibrato_env: ParamExpression,
        _note_list: list[SVNote],
        _pitch_controls: list[SVPitchControl] | None,
    ) -> None:
        self.synchronizer = _synchronizer
        self.pitch_diff = _pitch_diff
        self.vibrato_env = _vibrato_env
        if not len(_note_list):
            return
        note_structs = [
            NoteStruct(
                key=sv_note.pitch,
                start=_synchronizer.get_actual_secs_from_ticks(position_to_ticks(sv_note.onset)),
                end=_synchronizer.get_actual_secs_from_ticks(
                    position_to_ticks(sv_note.onset + sv_note.duration)
                ),
                portamento_offset=sv_note.attributes.transition_offset,
                portamento_left=sv_note.attributes.portamento_left,
                portamento_right=sv_note.attributes.portamento_right,
                depth_left=sv_note.attributes.depth_left,
                depth_right=sv_note.attributes.depth_right,
                vibrato_start=sv_note.attributes.vibrato_start,
                vibrato_left=sv_note.attributes.vibrato_left,
                vibrato_right=sv_note.attributes.vibrato_right,
                vibrato_depth=sv_note.attributes.vibrato_depth,
                vibrato_frequency=sv_note.attributes.vibrato_frequency,
                vibrato_phase=sv_note.attributes.vibrato_phase,
            )
            for sv_note in _note_list
        ]
        self.base_layer = BaseLayerGenerator(note_structs)
        self.vibrato_layer = VibratoLayerGenerator(note_structs)
        self.gaussian_layer = GaussianLayerGenerator(note_structs)
        if _pitch_controls:
            self.pitch_control_layer = PitchControlLayerGenerator(_pitch_controls)

    def value_at_ticks(self, ticks: int) -> float:
        return self.value_at_secs(self.synchronizer.get_actual_secs_from_ticks(ticks))

    def value_at_secs(self, secs: float) -> float:
        ticks = round(self.synchronizer.get_actual_ticks_from_secs(secs))
        if self.pitch_control_layer:
            pitch_control_value = self.pitch_control_layer.value_at_ticks(ticks)
        else:
            pitch_control_value = None
        if pitch_control_value is None:
            pitch_control_value = (
                self.base_layer.pitch_at_secs(secs)
                + self.vibrato_layer.pitch_diff_at_secs(secs)
                + self.gaussian_layer.pitch_diff_at_secs(secs)
            )
        return (
            pitch_control_value
            + self.pitch_diff.value_at_ticks(ticks) * self.vibrato_env.value_at_ticks(ticks) / 1000
        )



import math

from construct import (
    Byte,
    Bytes,
    BytesInteger,
    Const,
    ExprAdapter,
    FixedSized,
    GreedyRange,
    Int16ub,
    Int16ul,
    Int32ub,
    Mapping,
    PaddedString,
    PascalString,
    Prefixed,
    PrefixedArray,
    Select,
    Struct,
    Subconstruct,
    Switch,
    this,
)
from construct_typed import Context

Int32ul = BytesInteger(4, swapped=True)


def ppsf_int_encoder(value: int, ctx: Context) -> int:
    low = value & 0x7F
    high = (value >> 7) << 8
    width = math.floor(high.bit_length() / 8)
    base = 0x80 << (width * 8)
    high += base
    return high + low


def ppsf_int_decoder(value: int, ctx: Context) -> int:
    high = value >> 8
    low = value & 0xFF
    width = math.floor(high.bit_length() / 8)
    base = 0x80 << (width * 8)
    high -= base
    return (high << 7) + low


def ppsf_prefixed_array(subcon: Subconstruct) -> Select:
    return Select(
        PrefixedArray(Byte, subcon),
        PrefixedArray(
            ExprAdapter(
                Int16ub,
                encoder=ppsf_int_encoder,
                decoder=ppsf_int_decoder,
            ),
            subcon,
        ),
        PrefixedArray(
            ExprAdapter(
                BytesInteger(3),
                encoder=ppsf_int_encoder,
                decoder=ppsf_int_decoder,
            ),
            subcon,
        ),
        PrefixedArray(
            ExprAdapter(
                Int32ub,
                encoder=ppsf_int_encoder,
                decoder=ppsf_int_decoder,
            ),
            subcon,
        ),
    )


PpsfTrackTags = Mapping(
    Bytes(4),
    {
        "AudioEventTrack": b"AETK",
        "AudioInDeviceTrack": b"AIDT",
        "AudioOutDeviceTrack": b"AODT",
        "AudioSendEventTrack": b"ASTR",
        "MidiEventTrack": b"METK",
        "MidiInDeviceTrack": b"MIDT",
        "MidiOutDeviceTrack": b"MODT",
        "MusicalParamTrack": b"MPTR",
        "Vocaloid3EventTrack": b"V3TK",
        "Vocaloid2EventTrack": b"VETK",
    },
)


PpsfTrack = Struct(
    "magic" / PpsfTrackTags,
    "size" / Int32ul,
    "index" / Int16ul,
    "data" / Bytes(this.size - 2),
)


PpsfClipTags = Mapping(
    Bytes(4),
    {
        "Vocaloid3NoteClip": b"V3CL",
        "VsqNoteClip": b"VNCL",
        "MidiNoteClip": b"MNCL",
        "MidiMetaClip": b"MMCL",
        "MidiSysExClip": b"MXCL",
        "AudioClip": b"ADCL",
        "AutomationClip": b"AMCL",
    },
)


PpsfClip = Struct(
    "magic" / PpsfClipTags,
    "size" / Int32ul,
    "index" / Int16ul,
    "data" / Bytes(this.size - 2),
)


PpsfPluginTags = Mapping(
    Bytes(4),
    {
        "Vocaloid3EventControlPlugin": b"E3PG",
        "AudioEventControlPlugin": b"EAPG",
        "MidiEventControlPlugin": b"EMPG",
        "MusicParamEventControlPlugin": b"EMPP",
        "SynthEventControlPlugin": b"ESPG",
        "VsqEventControlPlugin": b"EVPG",
        "OpenVocaloidSynthPlugin": b"OSPG",
        "Vocaloid3SynthEventControlPlugin": b"V3CP",
        "VstAudioPlugin": b"VAPG",
        "VstMidiPlugin": b"VMPG",
        "VstSynthPlugin": b"VSPG",
    },
)


PpsfPlugin = Struct(
    "magic" / PpsfPluginTags,
    "size" / Int32ul,
    "index" / Int16ul,
    "data" / Bytes(this.size - 2),
)

PpsfTracksTags = Mapping(
    Bytes(4),
    {
        "AudioEventTracks": b"AETS",
        "AudioInDeviceTracks": b"AIDS",
        "AudioOutDeviceTracks": b"AODS",
        "AudioSendEventTracks": b"ASTS",
        "MidiEventTracks": b"METS",
        "MidiInDeviceTracks": b"MIDS",
        "MidiOutDeviceTracks": b"MODS",
        "MusicalParamTracks": b"MPTS",
        "Vocaloid3EventTracks": b"V3TS",
        "Vocaloid2EventTracks": b"VETS",
    },
)

PpsfTracks = Struct(
    "magic" / PpsfTracksTags,
    "size" / Int32ul,
    "data" / ppsf_prefixed_array(PpsfTrack),
)

PpsfEventTags = Mapping(
    Bytes(1),
    {
        "MidiEvent": b"\x00",
        "MidiSysExEvent": b"\x03",
        "MidiMetaEvent": b"\x04",
        "VsqNoteEvent": b"\x05",
        "MidiNoteEvent": b"\x07",
        "Vocaloid3NoteEvent": b"\x08",
        "AutomationEvent": b"\x09",
        "AudioEvent": b"\x0b",
    },
)

PpsfEvent = Struct(
    "magic" / PpsfEventTags,
    "size" / Int16ul,
    "data" / Bytes(this.size),
)

PpsfRect = Struct(
    "magic" / Const(b"RECT"),
    "size" / Int32ul,
    "width" / Int32ul,
    "height" / Int32ul,
    "x" / Int32ul,
    "y" / Int32ul,
)

PpsfEditorDataTags = Mapping(
    Bytes(4),
    {
        "EditorTrackData": b"ETRS",
        "EditorClipData": b"ECLS",
        "EditorNoteData": b"ENOT",
        "EditorEventData": b"EEVT",
    },
)

PpsfEditorData = Struct(
    "magic" / PpsfEditorDataTags,
    "size" / Int32ul,
    "data" / Bytes(this.size),
)

PpsfMarker = Struct(
    "magic" / Const(b"UMKR"),
    "size" / Int32ul,
    "data"
    / FixedSized(
        this.size,
        GreedyRange(
            Struct(
                "index" / Bytes(4),
                "tick" / Int32ul,
            )
        ),
    ),
)

PpsfEditorClipData = Struct(
    "magic" / Const(b"ECLS"),
    "size" / Int32ul,
    "data"
    / FixedSized(
        this.size,
        Struct(
            "unknown1" / Bytes(6),
            "unknown2" / PascalString(Byte, "utf-8"),
            "unknown3" / Bytes(13),
            "note_datas" / ppsf_prefixed_array(PpsfEditorData),
        ),
    ),
)

PpsfEditorTrackData = Struct(
    "magic" / Const(b"ETRS"),
    "size" / Int32ul,
    "data"
    / FixedSized(
        this.size,
        Struct(
            "unknown1" / Bytes(28),
            "unknown2" / PascalString(Byte, "utf-8"),
            "unknown3" / Bytes(22),
            "track_datas" / ppsf_prefixed_array(PpsfEditorData),
            "clip_datas" / ppsf_prefixed_array(PpsfEditorClipData),
            "event_datas" / ppsf_prefixed_array(PpsfEditorData),
        ),
    ),
)

PpsfChunkTags = Mapping(
    Bytes(4),
    {
        "Info": b"INFO",
        "Project": b"PROJ",
        "Transport": b"TRNS",
        "Config": b"CONF",
        "Devices": b"DVCS",
        "Tracks": b"TRKS",
        "Clips": b"CLPS",
        "Events": b"EVTS",
        "Plugins": b"PLGS",
        "EditorDatas": b"EDTS",
    },
)

PpsfChunk = Struct(
    "magic" / PpsfChunkTags,
    "size" / Int32ul,
    "data"
    / Switch(
        this.magic,
        {
            "Tracks": Struct("tracks" / GreedyRange(PpsfTracks), Bytes(1)),
            "Clips": Struct(
                "clips" / GreedyRange(ppsf_prefixed_array(PpsfClip)),
            ),
            "Events": Struct(
                "events" / GreedyRange(ppsf_prefixed_array(PpsfEvent)),
            ),
            "Plugins": Struct(
                "plugins" / GreedyRange(ppsf_prefixed_array(PpsfPlugin)),
            ),
            "EditorDatas": Struct(
                "prefix" / Bytes(4),
                "rect1" / PpsfRect,
                "unknown1" / Bytes(12),
                "unknown2" / PascalString(Byte, "utf-8"),
                "unknown3" / Bytes(48),
                "markers" / ppsf_prefixed_array(PpsfMarker),
                "unknown4" / ppsf_prefixed_array(PpsfEditorData),
                "editor_datas" / ppsf_prefixed_array(PpsfEditorTrackData),
                "other_track_datas" / GreedyRange(PpsfEditorData),
                "padding" / Bytes(1),
                "rect2" / PpsfRect,
                "rect3" / PpsfRect,
                "suffix" / Bytes(12),
            ),
        },
        Bytes(this.size),
    ),
)

PpsfLegacyProject = Struct(
    "magic" / Const("PPSF", PaddedString(4, "utf8")),
    "body"
    / Prefixed(
        Int32ul,
        Struct(
            "version" / PascalString(Int16ul, "utf8"),
            "chunks" / GreedyRange(PpsfChunk),
        ),
    ),
)



from collections.abc import Iterable

from libresvip.model.base import (
    ParamCurve,
    Params,
    Points,
    Project,
    SingingTrack,
    TimeSignature,
)
from libresvip.model.point import Point
from libresvip.model.reset_time_axis import reset_time_axis


def split_into_segments(
    project: Project, min_interval: int = 400, min_length: int = 5000
) -> Iterable[tuple[float, Project, float]]:
    track = next((t for t in project.track_list if isinstance(t, SingingTrack)), None)
    if not track or not track.note_list:
        return

    project = reset_time_axis(project)
    buffer = [track.note_list[0]]

    cur_seg_start = max(
        track.note_list[0].start_pos - 600,
        int(track.note_list[0].start_pos * 0.8),
    )
    cur_seg_interval = track.note_list[0].start_pos
    for i in range(1, len(track.note_list)):
        prev = track.note_list[i - 1]
        cur = track.note_list[i]
        interval = cur.start_pos - prev.end_pos
        if (
            interval >= min_interval
            and cur.start_pos - interval * 0.8 - cur_seg_start >= min_length
        ):
            prepare_space = min(600, int(cur_seg_interval * 0.8))
            trailing_space = min(400, int(cur_seg_interval * 0.2))
            seg_note_start_pos = buffer[0].start_pos
            pitch_points = [
                Point(pos - seg_note_start_pos + prepare_space, val)
                for pos, val in track.edited_params.pitch.points.root
                if 1920 <= pos - 1920 <= buffer[-1].end_pos + 50
            ]
            gender_points = [
                Point(pos - seg_note_start_pos + prepare_space, val)
                for pos, val in track.edited_params.gender.points.root
                if 1920 <= pos - 1920 <= buffer[-1].end_pos + 50
            ]
            for note in buffer:
                note.start_pos = note.start_pos - seg_note_start_pos + prepare_space

            cur_seg_start = cur.start_pos - min(600, int(interval * 0.8))
            cur_seg_interval = interval
            segment = Project(
                song_tempo_list=project.song_tempo_list,
                time_signature_list=[TimeSignature(bar_index=0, numerator=4, denominator=4)],
                track_list=[
                    SingingTrack(
                        note_list=buffer,
                        edited_params=Params(
                            pitch=ParamCurve(points=Points(root=pitch_points)),
                            gender=ParamCurve(points=Points(root=gender_points)),
                        ),
                    )
                ],
            )
            yield (
                (seg_note_start_pos - prepare_space) / 1000.0,
                segment,
                trailing_space / 1000.0,
            )

            buffer = []

        buffer.append(cur)

    prepare_space = min(600, int(cur_seg_interval * 0.8))
    seg_note_start_pos = buffer[0].start_pos
    pitch_points = [
        Point(pos - seg_note_start_pos + prepare_space, val)
        for pos, val in track.edited_params.pitch.points.root
        if 1920 <= pos - 1920 <= buffer[-1].end_pos + 50
    ]
    gender_points = [
        Point(pos - seg_note_start_pos + prepare_space, val)
        for pos, val in track.edited_params.gender.points.root
        if 1920 <= pos - 1920 <= buffer[-1].end_pos + 50
    ]
    for note in buffer:
        note.start_pos = note.start_pos - seg_note_start_pos + prepare_space

    segment = Project(
        song_tempo_list=project.song_tempo_list,
        time_signature_list=[TimeSignature(bar_index=0, numerator=4, denominator=4)],
        track_list=[
            SingingTrack(
                note_list=buffer,
                edited_params=Params(
                    pitch=ParamCurve(points=Points(root=pitch_points)),
                    gender=ParamCurve(points=Points(root=gender_points)),
                ),
            )
        ],
    )
    yield (seg_note_start_pos - prepare_space) / 1000.0, segment, 0.5



import dataclasses
import re
from typing import Any
from urllib.parse import urljoin

import pypinyin
from google.protobuf import any_pb2

from libresvip.core.lyric_phoneme.chinese import CHINESE_RE
from libresvip.core.tick_counter import skip_tempo_list
from libresvip.core.time_sync import TimeSynchronizer
from libresvip.core.warning_types import show_warning
from libresvip.model.base import (
    InstrumentalTrack,
    Note,
    ParamCurve,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
    Track,
)
from libresvip.utils.audio import audio_track_info
from libresvip.utils.music_math import ratio_to_db
from libresvip.utils.translation import gettext_lazy as _

from .color_pool import random_color
from .constants import (
    DEFAULT_SINGER_ID,
    MAX_NOTE_DURATION,
    MIN_NOTE_DURATION,
    TYPE_URL_BASE,
    Svip3TrackType,
)
from .model import (
    Svip3AudioPattern,
    Svip3AudioTrack,
    Svip3BeatSize,
    Svip3LineParamNode,
    Svip3Note,
    Svip3Project,
    Svip3SingingPattern,
    Svip3SingingTrack,
    Svip3SongBeat,
    Svip3SongTempo,
)
from .options import OutputOptions
from .singers import singers_data

PINYIN_PATTERN = re.compile(r"^[a-z]+$")


@dataclasses.dataclass
class Svip3Generator:
    options: OutputOptions
    first_bar_length: int = dataclasses.field(init=False)
    song_tempo_list: list[SongTempo] = dataclasses.field(init=False)
    synchronizer: TimeSynchronizer = dataclasses.field(init=False)
    song_duration: int = dataclasses.field(default=0)

    def generate_project(self, project: Project) -> Svip3Project:
        self.first_bar_length = round(project.time_signature_list[0].bar_length())
        self.synchronizer = TimeSynchronizer(project.song_tempo_list)
        return Svip3Project(
            beat_list=self.generate_time_signatures(project.time_signature_list),
            tempo_list=self.generate_song_tempos(project.song_tempo_list),
            track_list=self.generate_tracks(project.track_list),
            duration=self.song_duration,
        )

    @staticmethod
    def generate_time_signatures(
        time_signature_list: list[TimeSignature],
    ) -> list[Svip3SongBeat]:
        return [
            Svip3SongBeat(
                beat_size=Svip3BeatSize(
                    numerator=time_signature.numerator,
                    denominator=time_signature.denominator,
                ),
                pos=time_signature.bar_index,
            )
            for time_signature in time_signature_list
        ]

    def generate_song_tempos(self, song_tempo_list: list[SongTempo]) -> list[Svip3SongTempo]:
        song_tempo_list = skip_tempo_list(song_tempo_list, self.first_bar_length)
        return [
            Svip3SongTempo(
                tempo=round(song_tempo.bpm * 100),
                pos=song_tempo.position,
            )
            for song_tempo in song_tempo_list
        ]

    def generate_tracks(self, track_list: list[Track]) -> list[any_pb2.Any]:
        svip3_track_list = []
        for track in track_list:
            color = random_color()
            if isinstance(track, SingingTrack):
                svip3_track = self.generate_singing_track(track, color)
                type_url = urljoin(TYPE_URL_BASE, Svip3TrackType.SINGING_TRACK)
            elif isinstance(track, InstrumentalTrack):
                svip3_track = self.generate_instrumental_track(track, color)
                type_url = urljoin(TYPE_URL_BASE, Svip3TrackType.AUDIO_TRACK)
            else:
                continue
            svip3_track_container = any_pb2.Any(
                type_url=type_url,
                value=type(svip3_track).serialize(svip3_track),
            )
            svip3_track_list.append(svip3_track_container)
        return svip3_track_list

    def generate_instrumental_track(self, track: InstrumentalTrack, color: str) -> Svip3AudioTrack:
        return Svip3AudioTrack(
            name=track.title,
            color=color,
            mute=track.mute,
            solo=track.solo,
            volume=self.to_decibel_volume(track.volume),
            pan=self.generate_pan(track.pan),
            pattern_list=self.generate_audio_patterns(track),
        )

    @staticmethod
    def to_decibel_volume(volume: float) -> float:
        return max(ratio_to_db(max(volume, 0.01)), -70) if volume > 0 else -70

    @staticmethod
    def generate_pan(pan: float) -> float:
        return pan * 10.0

    def generate_audio_patterns(self, track: InstrumentalTrack) -> list[Svip3AudioPattern]:
        kwargs: dict[str, Any] = {}
        if (track_info := audio_track_info(track.audio_file_path)) is not None:
            audio_duration_in_ticks = round(
                self.synchronizer.get_actual_ticks_from_secs(track_info.duration / 1000)
            )
            kwargs["real_dur"] = kwargs["play_dur"] = audio_duration_in_ticks
            if track.offset >= 0:
                kwargs["play_pos"] = 0
                kwargs["real_pos"] = round(
                    self.synchronizer.get_actual_ticks_from_ticks(track.offset)
                )
            else:
                kwargs["play_pos"] = kwargs["real_dur"] - track.offset
                kwargs["play_pos"] = track.offset
                kwargs["real_pos"] = -track.offset
            pattern_end = kwargs["real_pos"] + kwargs["play_dur"]
            if pattern_end > self.song_duration:
                self.song_duration = round(pattern_end + self.first_bar_length)
        patterns = []
        if kwargs:
            kwargs |= {
                "audio_file_path": track.audio_file_path,
                "name": track.title,
            }
            patterns.append(Svip3AudioPattern(**kwargs))
        return patterns

    def generate_singing_track(self, track: SingingTrack, color: str) -> Svip3SingingTrack:
        return Svip3SingingTrack(
            name=track.title,
            color=color,
            mute=track.mute,
            solo=track.solo,
            volume=self.to_decibel_volume(track.volume),
            pan=self.generate_pan(track.pan),
            ai_singer_id=singers_data.inverse.get(track.ai_singer_name, DEFAULT_SINGER_ID),
            pattern_list=self.generate_singing_patterns(track),
        )

    def generate_singing_patterns(self, track: SingingTrack) -> list[Svip3SingingPattern]:
        if not track.note_list:
            return []
        last_note = track.note_list[-1]
        if last_note.end_pos > self.song_duration:
            self.song_duration = last_note.end_pos + self.first_bar_length
        return [
            Svip3SingingPattern(
                name=track.title,
                real_pos=0,
                play_pos=0,
                real_dur=last_note.end_pos + self.first_bar_length,
                play_dur=last_note.end_pos + self.first_bar_length,
                is_mute=track.mute,
                note_list=self.generate_notes(track.note_list),
                edited_pitch_line=self.generate_pitch_param(track.edited_params.pitch),
            )
        ]

    def generate_notes(self, note_list: list[Note]) -> list[Svip3Note]:
        svip3_note_list = []
        for note in note_list:
            consonant_length = 0
            has_consonant = False
            if note.edited_phones is not None and note.edited_phones.head_length_in_secs > 0:
                has_consonant = True
                phone_start_in_secs = (
                    self.synchronizer.get_actual_secs_from_ticks(note.start_pos)
                    - note.edited_phones.head_length_in_secs
                )
                phone_start_in_ticks = self.synchronizer.get_actual_ticks_from_secs(
                    phone_start_in_secs
                )
                consonant_length = round(
                    self.synchronizer.get_actual_ticks_from_ticks(note.start_pos)
                    - phone_start_in_ticks
                )
            if PINYIN_PATTERN.fullmatch(note.lyric) is not None:
                pronunciation = note.lyric
            elif note.pronunciation:
                pronunciation = note.pronunciation
            elif note.lyric and CHINESE_RE.fullmatch(note.lyric) is not None:
                pronunciation = " ".join(pypinyin.lazy_pinyin(note.lyric))
            else:
                pronunciation = note.pronunciation or note.lyric
            if note.lyric != "-" and PINYIN_PATTERN.fullmatch(pronunciation) is None:
                msg_prefix = _("Unsupported pinyin:")
                show_warning(f"{msg_prefix} {pronunciation}")
                pronunciation = ""
            note_duration = self.synchronizer.get_duration_secs_from_ticks(
                note.start_pos, note.end_pos
            )
            if note_duration < MIN_NOTE_DURATION:
                msg_prefix = _("Note duration is too short:")
                show_warning(f"{msg_prefix} {note.lyric}")
            elif note_duration > MAX_NOTE_DURATION:
                msg_prefix = _("Note duration is too long:")
                show_warning(f"{msg_prefix} {note.lyric}")
            svip3_note = Svip3Note(
                start_pos=note.start_pos,
                width_pos=note.length,
                key_index=note.key_number,
                lyric=note.lyric,
                pronouncing=pronunciation,
                consonant_len=consonant_length,
                has_consonant=has_consonant,
                sp_len=400 if note.head_tag == "V" else 0,
                sil_len=400 if note.head_tag == "0" else 0,
            )
            svip3_note_list.append(svip3_note)
        return svip3_note_list

    def generate_pitch_param(self, curve: ParamCurve) -> list[Svip3LineParamNode]:
        point_list = []
        splited_curves = curve.split_into_segments(-100)
        for segment in splited_curves:
            if not len(segment):
                continue
            prev_pos = 0
            left_interrupt = False
            for point in segment:
                if point.x == prev_pos:
                    continue
                if point.x < self.first_bar_length:
                    continue
                elif not left_interrupt:
                    left_interrupt = True
                    point_list.append(
                        Svip3LineParamNode(
                            pos=point.x - self.first_bar_length,
                            value=-1.0,
                        )
                    )
                pos = point.x - self.first_bar_length
                value = (point.y + 50) / 100
                point_list.append(
                    Svip3LineParamNode(
                        pos=pos,
                        value=float(value),
                    )
                )
                prev_pos = point.x
            last_point = segment[-1]
            point_list.append(
                Svip3LineParamNode(
                    pos=last_point.x - self.first_bar_length,
                    value=-1.0,
                )
            )
        return point_list



import dataclasses
import math

import more_itertools

from libresvip.core.tick_counter import skip_tempo_list
from libresvip.core.time_sync import TimeSynchronizer
from libresvip.model.base import (
    InstrumentalTrack,
    Note,
    ParamCurve,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
    Track,
)
from libresvip.utils.audio import audio_track_info
from libresvip.utils.music_math import ratio_to_db

from .model import (
    TuneLabAudioPart,
    TuneLabMidiPart,
    TuneLabNote,
    TuneLabPoint,
    TuneLabPoints,
    TuneLabProject,
    TuneLabTempo,
    TuneLabTimeSignature,
    TuneLabTrack,
)
from .options import OutputOptions


@dataclasses.dataclass
class TuneLabGenerator:
    options: OutputOptions
    synchronizer: TimeSynchronizer = dataclasses.field(init=False)
    first_bar_length: int = dataclasses.field(init=False)

    def generate_project(self, project: Project) -> TuneLabProject:
        self.synchronizer = TimeSynchronizer(project.song_tempo_list)
        self.first_bar_length = round(project.time_signature_list[0].bar_length())
        tempo_list = self.generate_tempos(project.song_tempo_list)
        time_signature_list = self.generate_time_signatures(project.time_signature_list)
        return TuneLabProject(
            tempos=tempo_list,
            time_signatures=time_signature_list,
            tracks=self.generate_tracks(project.track_list),
        )

    def generate_tempos(self, song_tempo_list: list[SongTempo]) -> list[TuneLabTempo]:
        song_tempo_list = skip_tempo_list(song_tempo_list, self.first_bar_length)
        return [
            TuneLabTempo(
                pos=tempo.position,
                bpm=tempo.bpm,
            )
            for tempo in song_tempo_list
        ]

    @staticmethod
    def generate_time_signatures(
        time_signature_list: list[TimeSignature],
    ) -> list[TuneLabTimeSignature]:
        return [
            TuneLabTimeSignature(
                numerator=time_signature.numerator,
                denominator=time_signature.denominator,
                bar_index=time_signature.bar_index,
            )
            for time_signature in time_signature_list
        ]

    def generate_tracks(self, track_list: list[Track]) -> list[TuneLabTrack]:
        tlp_track_list = []
        for track in track_list:
            if isinstance(track, InstrumentalTrack):
                if (track_info := audio_track_info(track.audio_file_path)) is not None:
                    tlp_track = TuneLabTrack(
                        name=track.title,
                        gain=self.generate_volume(track.volume),
                        pan=track.pan,
                        mute=track.mute,
                        solo=track.solo,
                        parts=[
                            TuneLabAudioPart(
                                name=track.title,
                                pos=track.offset,
                                path=track.audio_file_path,
                                dur=self.synchronizer.get_actual_ticks_from_secs_offset(
                                    track.offset, track_info.duration / 1000
                                ),
                            )
                        ],
                    )
                    tlp_track_list.append(tlp_track)
            elif isinstance(track, SingingTrack) and track.note_list:
                tlp_midi_part = TuneLabMidiPart(
                    name=track.title,
                    pos=0.0,
                    dur=math.ceil(track.note_list[-1].end_pos / self.first_bar_length)
                    * self.first_bar_length,
                    notes=self.generate_notes(track.note_list),
                )
                if pitch := self.generate_pitch(track.edited_params.pitch):
                    tlp_midi_part.pitch = pitch
                tlp_track = TuneLabTrack(
                    name=track.title,
                    gain=self.generate_volume(track.volume),
                    pan=track.pan,
                    mute=track.mute,
                    solo=track.solo,
                    parts=[tlp_midi_part],
                )
                tlp_track_list.append(tlp_track)
        return tlp_track_list

    @staticmethod
    def generate_volume(volume: float) -> float:
        return max(ratio_to_db(max(volume, 0.01)), -70) if volume > 0 else -70

    @staticmethod
    def generate_notes(note_list: list[Note]) -> list[TuneLabNote]:
        return [
            TuneLabNote(
                pos=note.start_pos,
                dur=note.length,
                pitch=note.key_number,
                lyric=note.lyric,
                pronunciation=note.pronunciation,
            )
            for note in note_list
        ]

    def generate_pitch(self, pitch: ParamCurve) -> list[TuneLabPoints]:
        return [
            TuneLabPoints(
                root=[
                    TuneLabPoint(
                        pos=point.x - self.first_bar_length,
                        value=point.y / 100,
                    )
                    for point in point_part
                ]
            )
            for point_part in more_itertools.split_at(
                pitch.points.root, lambda point: point.y == -100
            )
            if len(point_part)
        ]



import dataclasses

from libresvip.core.tick_counter import skip_tempo_list
from libresvip.core.time_sync import TimeSynchronizer
from libresvip.model.base import (
    InstrumentalTrack,
    Note,
    Params,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
    Track,
)
from libresvip.model.pitch_simulator import PitchSimulator
from libresvip.model.point import Point
from libresvip.model.portamento import PortamentoPitch
from libresvip.model.relative_pitch_curve import RelativePitchCurve
from libresvip.utils.music_math import ratio_to_db

from .model import (
    S5pInstrumental,
    S5pMeterItem,
    S5pMixer,
    S5pNote,
    S5pParameters,
    S5pPoint,
    S5pPoints,
    S5pProject,
    S5pTempoItem,
    S5pTrack,
    S5pTrackMixer,
)
from .options import OutputOptions

TICK_RATE = 1470000


@dataclasses.dataclass
class SynthVEditorGenerator:
    options: OutputOptions
    first_bar_length: int = dataclasses.field(init=False)
    synchronizer: TimeSynchronizer = dataclasses.field(init=False)

    def generate_project(self, project: Project) -> S5pProject:
        self.first_bar_length = round(project.time_signature_list[0].bar_length())
        s5p_project = S5pProject(
            tempo=self.generate_tempos(project.song_tempo_list),
            meter=self.generate_time_signatures(project.time_signature_list),
            tracks=self.generate_singing_tracks(project.track_list),
        )
        if (
            instrumental_track_and_mixer := next(
                (
                    self.generate_instrumental_track_and_mixer(track)
                    for track in project.track_list
                    if isinstance(track, InstrumentalTrack)
                ),
                None,
            )
        ) is not None:
            s5p_project.instrumental, s5p_project.mixer = instrumental_track_and_mixer
        return s5p_project

    def generate_tempos(self, song_tempo_list: list[SongTempo]) -> list[S5pTempoItem]:
        song_tempo_list = skip_tempo_list(song_tempo_list, self.first_bar_length)
        self.synchronizer = TimeSynchronizer(song_tempo_list)
        return [
            S5pTempoItem(
                position=song_tempo.position * TICK_RATE,
                beat_per_minute=song_tempo.bpm,
            )
            for song_tempo in song_tempo_list
        ]

    @staticmethod
    def generate_time_signatures(
        time_signature_list: list[TimeSignature],
    ) -> list[S5pMeterItem]:
        return [
            S5pMeterItem(
                measure=time_signature.bar_index,
                beat_per_measure=time_signature.numerator,
                beat_granularity=time_signature.denominator,
            )
            for time_signature in time_signature_list
        ]

    @staticmethod
    def generate_volume(volume: float) -> float:
        return max(ratio_to_db(max(volume, 0.01)), -70) if volume > 0 else -70

    def generate_singing_tracks(self, track_list: list[Track]) -> list[S5pTrack]:
        tracks = []
        for i, track in enumerate(track_list):
            if isinstance(track, SingingTrack):
                track_mixer = S5pTrackMixer(
                    solo=track.solo,
                    muted=track.mute,
                    pan=track.pan,
                    gain_decibel=self.generate_volume(track.volume),
                )

                s5p_track = S5pTrack(
                    display_order=i,
                    name=track.title,
                    db_name=track.ai_singer_name,
                    notes=self.generate_notes(track.note_list),
                    mixer=track_mixer,
                )
                if track.edited_params is not None:
                    s5p_track.parameters = self.generate_parameters(
                        track.edited_params, track.note_list
                    )
                tracks.append(s5p_track)
        if not tracks:
            tracks.append(
                S5pTrack(
                    name="Unnamed Track",
                    notes=[None],
                )
            )
        return tracks

    def generate_notes(self, note_list: list[Note]) -> list[S5pNote]:
        return [
            S5pNote(
                lyric=note.lyric,
                onset=note.start_pos * TICK_RATE,
                duration=note.length * TICK_RATE,
                pitch=note.key_number,
                d_f0_vbr=0,
            )
            for note in note_list
        ]

    def generate_instrumental_track_and_mixer(
        self, track: InstrumentalTrack
    ) -> tuple[S5pInstrumental, S5pMixer]:
        return S5pInstrumental(
            filename=track.audio_file_path,
            offset=self.synchronizer.get_actual_secs_from_ticks(track.offset),
        ), S5pMixer(
            gain_instrumental_decibel=self.generate_volume(track.volume),
            instrumental_muted=track.mute,
        )

    def generate_parameters(self, edited_params: Params, note_list: list[Note]) -> S5pParameters:
        interval = round(TICK_RATE * 3.75)
        pitch_simulator = PitchSimulator(
            synchronizer=self.synchronizer,
            portamento=PortamentoPitch.sigmoid_portamento(),
            note_list=note_list,
        )
        rel_pitch_points = RelativePitchCurve(self.first_bar_length).from_absolute(
            edited_params.pitch.points.root, pitch_simulator
        )
        return S5pParameters(
            pitch_delta=self.generate_pitch_delta(rel_pitch_points, interval),
            interval=interval,
        )

    def generate_pitch_delta(self, pitch: list[Point], interval: int) -> S5pPoints:
        return S5pPoints(
            root=[
                S5pPoint(
                    offset=round(point.x / (interval / TICK_RATE)),
                    value=point.y,
                )
                for point in pitch
                if point.y != -100
            ]
        )



import math

from libresvip.core.time_sync import TimeSynchronizer
from libresvip.model.base import Note
from libresvip.model.point import Point
from libresvip.utils.music_math import clamp

from .model import UtauNoteVibrato


def append_utau_note_vibrato(
    note_values: list[Point],
    vibrato_params: UtauNoteVibrato | None,
    note_start: Note,
    tick_time_transformer: TimeSynchronizer,
    sample_interval_tick: int,
) -> list[Point]:
    if vibrato_params is None:
        return note_values

    # x-axis: milliSec, y-axis: 100cents
    note_length = (
        tick_time_transformer.get_duration_secs_from_ticks(
            start_ticks=note_start.start_pos,
            end_ticks=note_start.end_pos,
        )
        * 1000
    )
    vibrato_length = note_length * vibrato_params.length / 100
    depth = vibrato_params.depth / 100
    if vibrato_length <= 0 or not vibrato_params.period or depth <= 0:
        return note_values
    frequency: float = 1.0 / vibrato_params.period
    ease_in_length = note_length * vibrato_params.fade_in / 100
    ease_out_length = note_length * vibrato_params.fade_out / 100
    phase = vibrato_params.phase_shift / 100
    shift = vibrato_params.shift / 100

    start = note_length - vibrato_length

    def vibrato(t: float) -> float:
        if t < start:
            return 0.0
        ease_in_factor = clamp((t - start) / ease_in_length, 0.0, 1.0) if ease_in_length else 1.0
        ease_out_factor = (
            clamp((note_length - t) / ease_out_length, 0.0, 1.0) if ease_out_length else 1.0
        )
        x = math.tau * (frequency * (t - start) - phase)
        return depth * ease_in_factor * ease_out_factor * (math.sin(x) + shift) * 100

    note_start_in_millis = (
        tick_time_transformer.get_actual_secs_from_ticks(note_start.start_pos) * 1000
    )

    # get approximate interval for interpolation
    sample_interval_in_millis = (
        tick_time_transformer.get_duration_secs_from_ticks(
            start_ticks=note_start.start_pos,
            end_ticks=note_start.start_pos + sample_interval_tick,
        )
        * 1000
    )

    interpolated_points = []
    for i in range(len(note_values)):
        x, y = note_values[i]
        x_in_millis = (
            tick_time_transformer.get_actual_secs_from_ticks(x) * 1000
        ) - note_start_in_millis
        interpolated_points.append((x_in_millis, y + vibrato(x_in_millis)))
        if i < len(note_values) - 1:
            next_x, _ = note_values[i + 1]
            pos = x_in_millis + sample_interval_in_millis
            while pos < (
                (tick_time_transformer.get_actual_secs_from_ticks(next_x) * 1000)
                - note_start_in_millis
            ):
                interpolated_points.append((pos, y + vibrato(pos)))
                pos += sample_interval_in_millis

    return [
        Point(
            round(
                tick_time_transformer.get_actual_ticks_from_secs((x + note_start_in_millis) / 1000)
            ),
            round(y),
        )
        for x, y in interpolated_points
    ]



# mypy: disable-error-code="misc"
from enum import Enum
from typing import Annotated

from pydantic import BaseModel, Field

from libresvip.utils.translation import gettext_lazy as _


class PronounciationConversionOptions(Enum):
    NONE: Annotated[str, Field(title=_("None"))] = "Do nothing"
    HANZI2PINYIN: Annotated[str, Field(title=_("Hanzi to Pinyin"))] = "Hanzi->Pinyin"
    TO_ROMAJI: Annotated[str, Field(title=_("To Romaji"))] = "->Romaji"
    TO_HIRAGANA: Annotated[str, Field(title=_("To Hiragana"))] = "->Hiragana"
    TO_KATAKANA: Annotated[str, Field(title=_("To Katakana"))] = "->Katakana"


class ProcessOptions(BaseModel):
    mode: PronounciationConversionOptions = Field(
        default=PronounciationConversionOptions.NONE,
        title=_("Pronounciation conversion mode"),
        description=_("Convert lyrics to the specified pronounciation format"),
    )



import dataclasses

import pypinyin

from libresvip.model.base import (
    Note,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
    Track,
)

from .model import (
    VogenNote,
    VogenProject,
    VogenTrack,
)
from .options import OutputOptions


@dataclasses.dataclass
class VogenGenerator:
    options: OutputOptions

    def generate_project(self, project: Project) -> VogenProject:
        return VogenProject(
            bpm0=self.generate_tempos(project.song_tempo_list),
            time_sig0=self.generate_time_signatures(project.time_signature_list),
            utts=self.generate_tracks(project.track_list),
        )

    def generate_tempos(self, song_tempo_list: list[SongTempo]) -> float:
        return song_tempo_list[0].bpm

    def generate_time_signatures(self, time_signature_list: list[TimeSignature]) -> str:
        time_signature = time_signature_list[0]
        return f"{time_signature.numerator}/{time_signature.denominator}"

    def generate_tracks(self, track_list: list[Track]) -> list[VogenTrack]:
        tracks = []
        for i, track in enumerate(track_list):
            if isinstance(track, SingingTrack):
                tracks.append(
                    VogenTrack(
                        name=track.title,
                        singer_id=track.ai_singer_name,
                        notes=self.generate_notes(track.note_list),
                    )
                )
        return tracks

    def generate_notes(self, note_list: list[Note]) -> list[VogenNote]:
        return [
            VogenNote(
                on=note.start_pos,
                dur=note.length,
                lyric=note.lyric,
                rom=note.pronunciation or " ".join(pypinyin.lazy_pinyin(note.lyric)),
                pitch=note.key_number,
            )
            for note in note_list
        ]



import dataclasses
import pathlib

from libresvip.core.constants import TICKS_IN_BEAT
from libresvip.core.time_sync import TimeSynchronizer
from libresvip.model.base import (
    InstrumentalTrack,
    Note,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
    Track,
)
from libresvip.model.point import Point

from .model import (
    VOXFactoryAudioTrack,
    VOXFactoryProject,
    VOXFactoryVocalClip,
)
from .options import InputOptions


@dataclasses.dataclass
class VOXFactoryParser:
    options: InputOptions
    path: pathlib.Path
    synchronizer: TimeSynchronizer = dataclasses.field(init=False)
    first_bar_length: int = dataclasses.field(init=False)

    def parse_project(self, vox_project: VOXFactoryProject) -> Project:
        return Project(
            song_tempo_list=self.parse_tempos(vox_project.tempo),
            time_signature_list=self.parse_time_signatures(vox_project.time_signature),
            track_list=self.parse_tracks(vox_project),
        )

    def parse_tempos(self, tempo: float) -> list[SongTempo]:
        tempos = [
            SongTempo(
                bpm=tempo,
                position=0,
            )
        ]
        self.synchronizer = TimeSynchronizer(tempos)
        return tempos

    def parse_time_signatures(self, time_signatures: list[int]) -> list[TimeSignature]:
        time_signature = TimeSignature(
            bar_index=0,
            numerator=time_signatures[0],
            denominator=time_signatures[1],
        )
        self.first_bar_length = int(time_signature.bar_length())
        return [time_signature]

    def parse_tracks(self, vox_project: VOXFactoryProject) -> list[Track]:
        tracks = []
        for track_name, track_data in sorted(
            vox_project.track_bank.items(),
            key=lambda x: vox_project.track_order.index(x[0]),
        ):
            if isinstance(track_data, VOXFactoryAudioTrack):
                if self.options.import_instrumental_track:
                    for _, clip_data in sorted(
                        track_data.clip_bank.items(),
                        key=lambda x: track_data.clip_order.index(x[0]),
                    ):
                        audio_path = (self.path.parent / clip_data.name).with_suffix(
                            pathlib.Path(clip_data.source_audio_data_key).suffix
                        )
                        if audio_path.exists():
                            track = InstrumentalTrack(
                                title=clip_data.name,
                                solo=track_data.solo,
                                mute=track_data.mute,
                                pan=track_data.pan,
                                offset=int(
                                    TICKS_IN_BEAT
                                    * (clip_data.start_quarter - clip_data.offset_quarter)
                                ),
                                audio_file_path="",
                            )
                            tracks.append(track)
            else:
                note_list = []
                pitch_points = []
                for _, clip_data in sorted(
                    track_data.clip_bank.items(),
                    key=lambda x: track_data.clip_order.index(x[0]),
                ):
                    clip_offset = int(
                        TICKS_IN_BEAT * (clip_data.start_quarter - clip_data.offset_quarter)
                    )
                    clip_notes, clip_pitch_points = self.parse_notes(clip_data, clip_offset)
                    note_list.extend(clip_notes)
                    pitch_points.extend(clip_pitch_points)
                if not note_list:
                    continue
                track = SingingTrack(
                    title=track_name,
                    solo=track_data.solo,
                    mute=track_data.mute,
                    pan=track_data.pan,
                    note_list=note_list,
                )
                if self.options.import_pitch:
                    pitch_points.insert(0, Point.start_point())
                    pitch_points.append(Point.end_point())
                    track.edited_params.pitch.points.root = pitch_points
                tracks.append(track)
        return tracks

    def parse_notes(
        self, clip_data: VOXFactoryVocalClip, clip_offset: int
    ) -> tuple[list[Note], list[Point]]:
        notes: list[Note] = []
        pitch_points: list[Point] = []
        f0_secs_step = 1024 / 44100
        for _, note_data in sorted(
            clip_data.note_bank.items(),
            key=lambda x: clip_data.note_order.index(x[0]),
        ):
            note = Note(
                start_pos=int(note_data.ticks + clip_offset),
                length=int(note_data.duration_ticks),
                key_number=note_data.midi,
                lyric=note_data.name,
                pronunciation=note_data.syllable,
            )
            if notes and note.start_pos < notes[-1].end_pos:
                notes[-1].length = note.start_pos - notes[-1].start_pos
            if self.options.import_pitch and note_data.pitch_bends:
                pitch_start = int(note_data.ticks + clip_offset + (note_data.pre_bend or 0))
                note_pitch_points = [Point(x=pitch_start + self.first_bar_length, y=-100)]
                pitch_start_secs = self.synchronizer.get_actual_secs_from_ticks(pitch_start)
                pitch_end_secs = pitch_start_secs + f0_secs_step * len(note_data.pitch_bends)
                pitch_end = self.synchronizer.get_actual_ticks_from_secs(pitch_end_secs)
                pitch_step = (pitch_end - pitch_start) / len(note_data.pitch_bends)
                for i, pitch_bend in enumerate(note_data.pitch_bends):
                    note_pitch_points.append(
                        Point(
                            x=int(pitch_start + i * pitch_step) + self.first_bar_length,
                            y=int((note_data.midi + pitch_bend) * 100),
                        )
                    )
                note_pitch_points.append(Point(x=int(pitch_end) + self.first_bar_length, y=-100))
                pitch_points.extend(note_pitch_points)
            notes.append(note)
        return notes, pitch_points



from contextvars import ContextVar

from libresvip.core.lyric_phoneme.chinese import get_pinyin_series
from libresvip.utils.translation import gettext_lazy as _

from .phoneme_dict import get_opencpop_dict

pinyin_list_ctx: ContextVar[list[str]] = ContextVar("pinyin_list")
phoneme_table_ctx: ContextVar[dict[str, str]] = ContextVar("phoneme_table")


def split(pinyin: str) -> tuple[str, str]:
    phoneme_table = phoneme_table_ctx.get()
    if pinyin not in phoneme_table:
        raise ValueError(
            _(
                'The selected dictionary does not contain the pronunciation "{}". Please check the pronunciation or try another dictionary.'
            ).format(pinyin)
        )

    phonemes = phoneme_table[pinyin].split()
    return ("", phonemes[0]) if len(phonemes) < 2 else (phonemes[0], phonemes[1])


def add_pinyin_from_lyrics(lyric_list: list[str]) -> None:
    pinyin_array = get_pinyin_series(lyric_list)
    pinyin_list = pinyin_list_ctx.get()
    pinyin_list.extend(pinyin_array)


def load_phoneme_table(dict_name: str) -> None:
    phoneme_table_ctx.set(get_opencpop_dict(dict_name))


def clear_all_pinyin() -> None:
    pinyin_list_ctx.set([])


def get_note_pinyin(note_lyric: str, index: int) -> str:
    pinyin_list = pinyin_list_ctx.get()
    return "-" if "-" in note_lyric else pinyin_list[index]



import dataclasses
import operator
from typing import cast

import more_itertools
from wanakana import PROLONGED_SOUND_MARK

from libresvip.core.tick_counter import skip_beat_list, skip_tempo_list
from libresvip.core.time_sync import TimeSynchronizer
from libresvip.model.base import (
    InstrumentalTrack,
    Note,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
    Track,
)

from .cevio_pitch import (
    CeVIOParamEvent,
    CeVIOTrackPitchData,
    pitch_from_cevio_track,
)
from .constants import OCTAVE_OFFSET, TICK_RATE
from .model import CeVIOCreativeStudioProject, CeVIOData, CeVIOGroup, CeVIOUnit
from .options import InputOptions


@dataclasses.dataclass
class CeVIOParser:
    options: InputOptions
    time_synchronizer: TimeSynchronizer = dataclasses.field(init=False)
    singer_id2name: dict[str | None, str] = dataclasses.field(default_factory=dict)

    def parse_project(self, ccs_project: CeVIOCreativeStudioProject) -> Project:
        scene_node = ccs_project.sequence.scene
        for sound_source in ccs_project.generation.svss.sound_sources.sound_source:
            if sound_source.name is not None:
                self.singer_id2name[sound_source.sound_source_id] = sound_source.name
        singing_unit_nodes = [
            unit_node for unit_node in scene_node.units.unit if unit_node.category == "SingerSong"
        ]
        audio_unit_nodes = [
            unit_node for unit_node in scene_node.units.unit if unit_node.category == "OuterAudio"
        ]
        id2group = {group.group_id: group for group in scene_node.groups.group}

        tracks: list[Track] = []
        tempos = []
        time_signatures = []
        for index, unit_node in enumerate(singing_unit_nodes):
            group_id = unit_node.group
            if (group := id2group.get(group_id)) is None:
                group = CeVIOGroup()
            track_name = group.name
            track, tempo_part, time_signature_part = self.parse_singing_track(
                index, unit_node, group, track_name
            )
            tracks.append(track)
            tempos.extend(tempo_part)
            time_signatures.extend(time_signature_part)

        tempos = self.merge_tempos(tempos)
        time_signatures = self.merge_time_signatures(time_signatures)

        self.time_synchronizer = TimeSynchronizer(tempos)

        if self.options.import_instrumental_track:
            for index, unit_node in enumerate(audio_unit_nodes):
                group_id = unit_node.group
                if (group := id2group.get(group_id)) is None:
                    group = CeVIOGroup()
                track_name = group.name
                track = self.parse_instrumental_track(index, unit_node, group, track_name)
                tracks.append(track)

        return Project(
            track_list=tracks,
            time_signature_list=skip_beat_list(time_signatures, 0),
            song_tempo_list=skip_tempo_list(tempos, 0),
        )

    def merge_tempos(self, tempos: list[SongTempo]) -> list[SongTempo]:
        buckets = more_itertools.bucket(tempos, operator.attrgetter("position"))
        return [next(buckets[key]) for key in buckets] or [SongTempo()]

    def merge_time_signatures(self, time_signatures: list[TimeSignature]) -> list[TimeSignature]:
        buckets = more_itertools.bucket(time_signatures, operator.attrgetter("bar_index"))
        return [next(buckets[key]) for key in buckets] or [TimeSignature()]

    def parse_instrumental_track(
        self,
        index: int,
        unit_node: CeVIOUnit,
        group_node: CeVIOGroup,
        track_name: str | None,
    ) -> InstrumentalTrack:
        return InstrumentalTrack(
            title=track_name or f"Track {index + 1}",
            audio_file_path=unit_node.file_path,
            mute=group_node.is_muted,
            solo=group_node.is_solo,
            offset=int(
                self.time_synchronizer.get_actual_ticks_from_secs(unit_node.start_time.duration)
            )
            if unit_node.start_time is not None
            else 0,
        )

    def parse_singing_track(
        self,
        index: int,
        unit_node: CeVIOUnit,
        group_node: CeVIOGroup,
        track_name: str | None,
    ) -> tuple[SingingTrack, list[SongTempo], list[TimeSignature]]:
        time_nodes = unit_node.song.beat.time if unit_node.song.beat is not None else []

        prev_tick = 0.0
        time_signatures = [
            TimeSignature(bar_index=0, numerator=4, denominator=4),
        ]
        for time_node in time_nodes:
            tick = time_node.clock // TICK_RATE
            numerator = time_node.beats
            denominator = time_node.beat_type

            if numerator is not None and denominator is not None:
                ticks_in_measure = time_signatures[-1].bar_length()
                tick_diff = tick - prev_tick
                measure_diff = tick_diff / ticks_in_measure
                time_signatures.append(
                    TimeSignature(
                        bar_index=int(
                            time_signatures[-1].bar_index + measure_diff,
                        ),
                        numerator=numerator,
                        denominator=denominator,
                    )
                )
                prev_tick = tick

        time_signatures = [
            time_signature.model_copy(update={"bar_index": time_signature.bar_index - 4})
            for time_signature in time_signatures
        ]

        tick_prefix = int(time_signatures[0].bar_length())

        tempo_nodes = unit_node.song.tempo.sound if unit_node.song.tempo is not None else []
        tempos = [
            SongTempo(
                position=tempo_node.clock // TICK_RATE,
                bpm=float(tempo_node.tempo),
            )
            for tempo_node in tempo_nodes
            if tempo_node.clock is not None and tempo_node.tempo is not None
        ]

        notes = []
        note_nodes = unit_node.song.score.note
        for note_node in note_nodes:
            tick_on = (note_node.clock // TICK_RATE) if note_node.clock is not None else None
            duration = (note_node.duration // TICK_RATE) if note_node.duration is not None else None
            pitch_step = note_node.pitch_step if note_node.pitch_step is not None else None
            pitch_octave = (
                note_node.pitch_octave - OCTAVE_OFFSET
                if note_node.pitch_octave is not None
                else None
            )
            key = (
                pitch_step + pitch_octave * 12
                if pitch_step is not None and pitch_octave is not None
                else None
            )
            lyric = note_node.lyric or None
            if (
                tick_on is not None
                and duration is not None
                and key is not None
                and lyric is not None
            ):
                if lyric == chr(PROLONGED_SOUND_MARK):
                    lyric = "-"
                phoneme = None
                if note_node.phonetic is not None:
                    phoneme = note_node.phonetic.replace(",", " ")
                notes.append(
                    Note(
                        key_number=key,
                        lyric=lyric,
                        start_pos=tick_on,
                        length=duration,
                        pronunciation=phoneme,
                    )
                )

        cevio_track_pitch_data = None
        if self.options.import_pitch and unit_node.song.parameter.log_f0 is not None:
            pitch_data_nodes: list[CeVIOData] = cast(
                "list[CeVIOData]", unit_node.song.parameter.log_f0.data
            )
            vibrato_amplitude_nodes: list[CeVIOData] = cast(
                "list[CeVIOData]",
                unit_node.song.parameter.vib_amp.data
                if unit_node.song.parameter.vib_amp is not None
                else [],
            )
            vibrato_frequency_nodes: list[CeVIOData] = cast(
                "list[CeVIOData]",
                unit_node.song.parameter.vib_frq.data
                if unit_node.song.parameter.vib_frq is not None
                else [],
            )
            pitch_datas = [
                pitch_data
                for pitch_data_node in pitch_data_nodes
                if (pitch_data := self.parse_param_data(pitch_data_node))
            ]
            vibrato_amplitude_data = [
                vibrato_amplitude
                for vibrato_amplitude_node in vibrato_amplitude_nodes
                if (vibrato_amplitude := self.parse_param_data(vibrato_amplitude_node))
            ]
            vibrato_frequency_data = [
                vibrato_frequency
                for vibrato_frequency_node in vibrato_frequency_nodes
                if (vibrato_frequency := self.parse_param_data(vibrato_frequency_node))
            ]
            cevio_track_pitch_data = CeVIOTrackPitchData(
                events=pitch_datas,
                tempos=tempos,
                tick_prefix=tick_prefix,
                vibrato_amplitude_events=vibrato_amplitude_data,
                vibrato_frequency_events=vibrato_frequency_data,
            )

        track_name = track_name or f"Track {index + 1}"
        track = SingingTrack(
            ai_singer_name=self.singer_id2name.get(unit_node.cast_id, ""),
            title=track_name,
            note_list=notes,
            mute=group_node.is_muted,
            solo=group_node.is_solo,
        )
        if (
            cevio_track_pitch_data is not None
            and (pitch := pitch_from_cevio_track(cevio_track_pitch_data)) is not None
        ):
            track.edited_params.pitch = pitch
        return track, tempos, time_signatures

    @staticmethod
    def parse_param_data(data_element: CeVIOData) -> CeVIOParamEvent | None:
        value = float(data_element.value) if data_element.value is not None else None
        if value is not None:
            index = data_element.index or None
            repeat = data_element.repeat or None
            return CeVIOParamEvent(idx=index, repeat=repeat, value=value)



from ko_pron.ko_pron import decompose_syllable, romanise

romaji2korean_xsampa = {
    "": "",
    "a": "a",
    "ya": "ja",
    "eo": "7",
    "yeo": "j7",
    "ae": "e",
    "e": "e",
    "yae": "je",
    "ye": "je",
    "o": "o",
    "wa": "oa",
    "yo": "jo",
    "u": "u",
    "weo": "u7",
    "wae": "ue",
    "we": "ue",
    "oe": "ue",
    "wi": "ui",
    "eu": "M",
    "eui": "Mi",
    "i": "i",
    "g": "g",
    "gg": "g'",
    "n": "n",
    "d": "d",
    "dd": "d'",
    "r": "r",
    "l": "l",
    "m": "m",
    "b": "b",
    "bb": "b'",
    "s": "s",
    "ss": "s'",
    "j": "c",
    "jj": "c'",
    "ch": "ch",
    "k": "k",
    "t": "t",
    "p": "p",
    "h": "h",
    "ng": "Np",
    "gs": "gp",
    "nch": "np",
    "nh": "np",
    "kk": "g'",
    "tt": "d'",
    "pp": "b'",
    "wo": "a",
    "yu": "ju",
    "ui": "i",
    "lg": "gp",
    "lm": "mp",
    "lb": "rp",
    "ls": "rp",
    "lt": "rp",
    "lp": "bp",
    "lh": "rp",
    "ps": "bp",
}

romaji2korean_xsampa_final = {
    "k": "gp",
    "n": "np",
    "t": "dp",
    "l": "rp",
    "m": "mp",
    "p": "bp",
}


def hangul2xsampa(lyric: str) -> str:
    phonemes = []
    for each in decompose_syllable(lyric):
        if isinstance(each, dict):
            initial_romaji = romanise(each["initial"], "rr")
            vowel_romaji = romanise(each["vowel"], "rr")
            final_romaji = romanise(each["final"], "rr")
            initial_xsampa = romaji2korean_xsampa.get(initial_romaji, "")
            vowel_xsampa = romaji2korean_xsampa.get(vowel_romaji, "")
            final_xsampa = romaji2korean_xsampa.get(final_romaji, "")
            if initial_xsampa in ["s", "sh"] and vowel_xsampa.startswith(("i", "y", "j")):
                initial_xsampa = "sh" if initial_xsampa == "s" else "sh'"
            elif vowel_xsampa in ["s", "sh"] and final_xsampa.startswith(("i", "y", "j")):
                vowel_xsampa = "sh" if vowel_xsampa == "s" else "sh'"
            if final_xsampa in romaji2korean_xsampa_final:
                final_xsampa = romaji2korean_xsampa_final[final_xsampa]
            phonemes.append(f"{initial_xsampa} {vowel_xsampa} {final_xsampa}".strip())
    return " ".join(phonemes) if phonemes else "r a"



import dataclasses
import operator
from typing import Any

import more_itertools

from libresvip.core.constants import TICKS_IN_BEAT
from libresvip.core.time_sync import TimeSynchronizer
from libresvip.model.base import Project, SingingTrack, SongTempo, TimeSignature, Track

from .model import VxPitchData, VxPitchPoint, VxTimeBasedPitchSequence
from .options import OutputOptions

Container = dict[str, Any]


@dataclasses.dataclass
class VxBetaGenerator:
    options: OutputOptions
    synchronizer: TimeSynchronizer = dataclasses.field(init=False)
    first_bar_length: int = dataclasses.field(init=False)

    @property
    def tick_rate(self) -> float:
        return TICKS_IN_BEAT / self.options.ticks_per_beat

    def generate_project(self, project: Project) -> Container:
        self.first_bar_length = round(
            project.time_signature_list[0].bar_length(self.options.ticks_per_beat)
        )
        self.synchronizer = TimeSynchronizer(project.song_tempo_list)
        master_track_events = []
        master_track_events.extend(self.generate_tempos(project.song_tempo_list))
        master_track_events.extend(self.generate_time_signatures(project.time_signature_list))
        if tracks := self.generate_tracks(project.track_list):
            tracks[0]["events"].extend(master_track_events)
        self._convert_cumulative_to_delta(tracks)
        return {
            "ticks_per_beat": self.options.ticks_per_beat,
            "tracks": tracks,
        }

    @staticmethod
    def _convert_cumulative_to_delta(tracks: list[Container]) -> None:
        for track in tracks:
            tick = 0
            track["events"] = sorted(track["events"], key=operator.itemgetter("time"))
            for event in track["events"]:
                tick, event["time"] = event["time"], event["time"] - tick

    def generate_tempos(self, song_tempo_list: list[SongTempo]) -> list[Container]:
        return [
            {
                "tempo": tempo.bpm,
                "time": round(tempo.position / self.tick_rate),
            }
            for tempo in song_tempo_list
            if tempo.position >= 0
        ]

    def generate_time_signatures(
        self,
        time_signature_list: list[TimeSignature],
    ) -> list[Container]:
        events = []
        prev_ticks = 0
        prev_time_signature = None
        for time_signature in time_signature_list:
            if time_signature.bar_index >= 0:
                if prev_time_signature is not None:
                    prev_ticks += round(
                        (time_signature.bar_index - prev_time_signature.bar_index)
                        * prev_time_signature.bar_length(self.options.ticks_per_beat)
                    )
                events.append(
                    {
                        "numerator": time_signature.numerator,
                        "denominator": time_signature.denominator,
                        "time": prev_ticks,
                    }
                )
                prev_time_signature = time_signature
        return events

    def generate_tracks(self, tracks: list[Track]) -> list[Container]:
        return [
            mido_track
            for track in tracks
            if isinstance(track, SingingTrack)
            and (mido_track := self.generate_track(track)) is not None
        ]

    def generate_track(self, track: SingingTrack) -> Container | None:
        events: list[Container] = []
        encoded_title = track.title.encode().decode("latin-1")
        title_parts = []
        for is_first, is_last, i in more_itertools.mark_ends(range(0, len(encoded_title), 12)):
            if len(encoded_title) <= 12:
                title_parts.append(
                    {"name": encoded_title[i : i + 12].encode("latin-1").decode(), "seq_stat": 0x10}
                )
            elif is_first:
                title_parts.append(
                    {"name": encoded_title[i : i + 12].encode("latin-1").decode(), "seq_stat": 0x50}
                )
            elif is_last:
                title_parts.append(
                    {"name": encoded_title[i : i + 12].encode("latin-1").decode(), "seq_stat": 0xD0}
                )
            else:
                title_parts.append(
                    {"name": encoded_title[i : i + 12].encode("latin-1").decode(), "seq_stat": 0x90}
                )
        for i, note in enumerate(track.note_list):
            lyric = "l-aa"
            if note.lyric.isascii():
                lyric = note.lyric
            elif note.pronunciation and note.pronunciation.isascii():
                lyric = note.pronunciation
            if lyric:
                for is_first, is_last, offset in more_itertools.mark_ends(
                    range(-2, len(lyric), 12)
                ):
                    if len(lyric) <= 10:
                        events.append(
                            {
                                "type": "lyrics",
                                "seq_num": i,
                                "seq_stat": 0x10,
                                "text": lyric[:10],
                                "time": 0,
                            }
                        )
                    elif is_first:
                        events.append(
                            {
                                "type": "lyrics",
                                "seq_num": i,
                                "seq_stat": 0x50,
                                "text": lyric[:10],
                                "time": 0,
                            }
                        )
                    elif is_last:
                        events.append(
                            {
                                "type": "lyrics",
                                "seq_num": None,
                                "seq_stat": 0xD0,
                                "text": lyric[offset:],
                                "time": 0,
                            }
                        )
                    else:
                        events.append(
                            {
                                "type": "lyrics",
                                "seq_num": None,
                                "seq_stat": 0x90,
                                "text": lyric[offset : offset + 12],
                                "time": 0,
                            }
                        )
            events.extend(
                [
                    {
                        "type": "note_on",
                        "note": note.key_number,
                        "time": round(note.start_pos / self.tick_rate),
                        "velocity": 0x7F,
                        "on_data": i,
                    },
                    {
                        "type": "note_off",
                        "note": note.key_number,
                        "time": round(note.end_pos / self.tick_rate),
                        "velocity": 0x00,
                        "off_data": 0x00,
                    },
                ]
            )
        prev_point = None
        secs_step = 1 / 200
        pitch_data = VxPitchData(
            time_based_pitch_sequence=VxTimeBasedPitchSequence(
                time_frame_period_seconds=secs_step,
            )
        )
        for pitch_point in track.edited_params.pitch.points.root:
            if pitch_point.y != -100:
                pos = int(
                    self.synchronizer.get_actual_secs_from_ticks(
                        pitch_point.x - self.first_bar_length
                    )
                    / secs_step
                )
                if prev_point is not None:
                    pitch_data.time_based_pitch_sequence.pitch_sequence.extend(
                        [
                            VxPitchPoint(
                                position=interp_pos,
                                pitch=(
                                    prev_point.pitch
                                    + (pitch_point.y - 6900 - prev_point.pitch)
                                    * (
                                        (interp_pos - prev_point.position)
                                        / (pos - prev_point.position)
                                    )
                                )
                                if interp_pos != pos
                                else pitch_point.y - 6900,
                            )
                            for interp_pos in range(prev_point.position + 1, pos + 1)
                        ]
                    )
                else:
                    pitch_data.time_based_pitch_sequence.pitch_sequence.append(
                        VxPitchPoint(position=pos, pitch=pitch_point.y - 6900)
                    )
                prev_point = pitch_data.time_based_pitch_sequence.pitch_sequence[-1]
            else:
                prev_point = None
            if pitch_data.time_based_pitch_sequence.pitch_sequence:
                pitch_data.time_based_pitch_sequence.num_frames_overall_sequence = (
                    pitch_data.time_based_pitch_sequence.pitch_sequence[-1].position
                )
        pitch_data_json = pitch_data.model_dump_json(by_alias=True)
        for is_first, is_last, offset in more_itertools.mark_ends(
            range(0, len(pitch_data_json), 12)
        ):
            if len(pitch_data_json) <= 12:
                events.append(
                    {
                        "type": "metadata",
                        "text": pitch_data_json[:12],
                        "time": 0,
                    }
                )
            elif is_first:
                events.append(
                    {
                        "type": "metadata",
                        "seq_stat": 0x50,
                        "text": pitch_data_json[:12],
                        "time": 0,
                    }
                )
            elif is_last:
                events.append(
                    {
                        "type": "metadata",
                        "seq_stat": 0xD0,
                        "text": pitch_data_json[offset:],
                        "time": 0,
                    }
                )
            else:
                events.append(
                    {
                        "type": "metadata",
                        "seq_stat": 0x90,
                        "text": pitch_data_json[offset : offset + 12],
                        "time": 0,
                    }
                )
        if events:
            return {"title_parts": title_parts, "events": events}



import csv
import io
from importlib.resources import files

from libresvip.utils.translation import gettext_lazy as _


def get_opencpop_dict(dict_name: str, g2p: bool = True) -> dict[str, str]:
    opencpop_dict = {}
    dict_dir = files("libresvip.plugins.ds") / "dicts"
    if (dict_content := (dict_dir / f"{dict_name}.txt").read_text(encoding="utf-8")) is None:
        msg = _("Cannot find dict.")
        raise FileNotFoundError(msg)
    reader = csv.DictReader(
        io.StringIO(dict_content),
        delimiter="\t",
        fieldnames=["pinyin", "phone"],
    )
    for row in reader:
        if g2p:
            opencpop_dict[row["pinyin"]] = row["phone"]
        else:
            opencpop_dict[row["phone"]] = row["pinyin"]
    return opencpop_dict



class InvalidFileTypeError(Exception):
    pass


class UnsupportedProjectVersionError(InvalidFileTypeError):
    pass


class NoTrackError(Exception):
    pass


class NotesOverlappedError(Exception):
    pass


class ParamsError(Exception):
    pass
